from ape.exceptions import AccountsError, ApeException, ProviderError
from ape.types import AddressType


class StarknetEcosystemError(ApeException):
    """
    A general Starknet ecosystem error.
    """


class StarknetProviderError(ProviderError):
    """
    An error raised by the Starknet provider.
    """


class StarknetTokensError(ApeException):
    """
    An error raised by the Starknet tokens manager.
    """


class StarknetAccountsError(AccountsError):
    """
    An error raised by a Starknet account.
    """


class ContractTypeNotFoundError(StarknetEcosystemError):
    """
    An error raised when unable to locate a contract type.
    """

    def __init__(self, address: AddressType):
        super().__init__(f"Failed to find contract type for '{address}'")



import pytest

AMOUNT_0 = 10
AMOUNT_1 = 2**128 + 42


@pytest.fixture(scope="module")
def receipt(token_contract, second_account, account):
    return token_contract.fire_events(second_account, AMOUNT_0, AMOUNT_1, sender=account)


def test_decode_logs(receipt, token_contract, account, second_account):
    expected_sender = int(account.address, 16)
    expected_receiver = int(second_account.address, 16)
    transfer_logs = list(receipt.decode_logs(token_contract.Transfer))

    # TODO: Figure out why 1 extra strange Transfer event shows up (as of 0.5)

    assert transfer_logs

    mint_logs = list(receipt.decode_logs(token_contract.Mint))
    assert len(mint_logs) == 1
    log = mint_logs[-1]
    assert log.sender == expected_sender
    assert log.amount0 == AMOUNT_0
    assert log.amount1 == AMOUNT_1
    assert log.to == expected_receiver

    # Verify events emitted from imported function call show up in receipt.
    lib_event = token_contract.contract_type.events["MyEventLib_ParentEvent"]
    lib_logs = list(receipt.decode_logs(lib_event))
    assert len(lib_logs) == 1
    assert lib_logs[0].favorite_account == expected_receiver


def test_decode_logs_when_logs_from_other_contract(token_contract, token_user_contract, account):
    receipt = token_user_contract.fireTokenEvent(token_contract.address, sender=account)
    transfer_logs = list(receipt.decode_logs(token_contract.Transfer))
    mint_logs = list(receipt.decode_logs(token_contract.Mint))

    assert transfer_logs
    assert len(mint_logs) == 1
    assert mint_logs[-1].amount0 == 100
    assert mint_logs[-1].amount1 == 200

    # The caller address is the user token
    assert mint_logs[-1].sender == int(token_user_contract.address, 16)

    # Verify events emitted from imported function call show up in receipt.
    lib_event = token_contract.contract_type.events["MyEventLib_ParentEvent"]
    lib_logs = list(receipt.decode_logs(lib_event))
    assert len(lib_logs) == 1


def test_decode_logs_no_specify_abi(receipt, account, second_account):
    logs = list(receipt.decode_logs())
    expected_from = int(account.address, 16)
    expected_to = int(second_account.address, 16)
    assert len(logs) >= 2
    transfer_log = [
        x
        for x in logs
        if x.event_name == "Transfer" and x.from_ == expected_from and x.to == expected_to
    ][-1]
    assert transfer_log.value == AMOUNT_0

    mint_log = [x for x in logs if x.event_name == "Mint"][-1]
    assert mint_log.sender == expected_from
    assert mint_log.amount0 == AMOUNT_0
    assert mint_log.amount1 == AMOUNT_1
    assert mint_log.to == expected_to



from typing import List, Optional, Union, cast

import click
from ape.api.networks import LOCAL_NETWORK_NAME
from ape.cli import ape_cli_context, existing_alias_argument, non_existing_alias_argument
from ape.cli.options import ApeCliContextObject
from ape.logging import logger
from ape.utils import add_padding_to_strings
from eth_utils import is_hex, to_hex
from starkware.crypto.signature.signature import EC_ORDER
from starkware.starknet.definitions.fields import ContractAddressSalt

from ape_starknet.accounts import (
    BaseStarknetAccount,
    StarknetAccountContainer,
    StarknetAccountDeployment,
    StarknetKeyfileAccount,
)
from ape_starknet.ecosystems import NETWORKS
from ape_starknet.utils import (
    ARGENTX_ACCOUNT_CLASS_HASH,
    OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH,
    PLUGIN_NAME,
    to_int,
)


def _get_container(cli_ctx: ApeCliContextObject) -> StarknetAccountContainer:
    return cast(StarknetAccountContainer, cli_ctx.account_manager.containers[PLUGIN_NAME])


def class_hash_option(default: Optional[Union[str, int]] = None, required: bool = False):
    def callback(ctx, param, value) -> Optional[int]:
        if value is None:
            return None

        elif isinstance(value, int):
            return int(value)

        elif value.lower() == "openzeppelin":
            return OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH

        elif value.lower() in ("argentx", "argent-x"):
            return ARGENTX_ACCOUNT_CLASS_HASH

        return to_int(value)

    return click.option(
        "--class-hash",
        help="The class hash of the account contract.",
        required=required,
        callback=callback,
        default=default,
        type=str,
    )


def address_option():
    return click.option(
        "--address",
        help="The contract address of the account",
        callback=lambda ctx, param, value: ctx.obj.network_manager.starknet.decode_address(value)
        if value
        else None,
    )


def _network_callback(ctx, param, value, single: bool = False):
    parse = ctx.obj.network_manager.parse_network_choice

    if single and value == "starknet":
        raise click.BadOptionUsage("--network", "Must use single network.")

    elif value == "starknet":
        # Use all live networks.
        return [_validate_network(parse, n) for n in NETWORKS]

    if single and value is not None:
        return _validate_network(parse, value)

    elif single:
        return value

    elif value is not None:
        return [_validate_network(parse, n.strip()) for n in value.split(",")]

    return []


def network_option(required: bool = False, default: Optional[str] = None, single: bool = False):
    def callback(ctx, param, value):
        return _network_callback(ctx, param, value, single=single)

    return click.option(
        "--network",
        default=default,
        required=required,
        callback=callback,
    )


@click.group("accounts")
def accounts():
    """Manage Starknet accounts"""


def _salt_callback(ctx, param, value):
    return value or ContractAddressSalt.get_random_value()


@accounts.command()
@ape_cli_context()
@non_existing_alias_argument()
@class_hash_option(default=OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH)
@click.option(
    "--constructor-calldata", help="Comma separated list of calldata, default tries to be smart."
)
@click.option(
    "--salt", help="The default address salt to use when deploying.", callback=_salt_callback
)
def create(cli_ctx, alias, class_hash, constructor_calldata, salt):
    """Create an account keypair"""
    calldata = (
        [cli_ctx.starknet.decode_primitive_value(x) for x in constructor_calldata.split(",")]
        if constructor_calldata is not None
        else None
    )
    _get_container(cli_ctx).create_account(
        alias,
        class_hash=class_hash,
        constructor_calldata=calldata,
        salt=salt,
        allow_local_file_store=True,
    )
    cli_ctx.logger.success(f"Created account key-pair for alias '{alias}'.")


def _funder_callback(ctx, param, value):
    container = _get_container(ctx.obj)
    if not value:
        return None

    if value.isnumeric():
        # Useful for testing / local but not realistic.
        return container.genesis_test_accounts[int(value)]

    return container.load(value)


@accounts.command()
@ape_cli_context()
@click.argument("alias")
@network_option(default=LOCAL_NETWORK_NAME, single=True)
@click.option("--funder", help="Use another an account to help fund", callback=_funder_callback)
def deploy(cli_ctx, network, alias, funder):
    """Deploy an account"""
    with cli_ctx.network_manager.parse_network_choice(network):
        logger.info(f"Deploying account '{alias}' to network '{network}'.")
        container = _get_container(cli_ctx)
        account = container.load(alias)
        uses_keyfile = isinstance(account, StarknetKeyfileAccount)

        if uses_keyfile:
            account.unlock("Enter passphrase to deploy account")  # type: ignore[attr-defined]

        try:
            receipt = account.deploy_account(funder=funder)
            contract_address_styled = click.style(receipt.contract_address, bold=True)
            logger.success(f"Account successfully deployed to '{contract_address_styled}'.")
        finally:
            if uses_keyfile:
                account.lock()  # type: ignore[attr-defined]


@accounts.command("list")
@ape_cli_context()
def _list(cli_ctx):
    """List your Starknet accounts"""

    starknet_accounts: List[StarknetKeyfileAccount] = [
        x for x in _get_container(cli_ctx).accounts if isinstance(x, StarknetKeyfileAccount)
    ]

    if len(starknet_accounts) == 0:
        cli_ctx.logger.info("No accounts found.")
        return

    num_accounts = len(starknet_accounts)
    header = f"Found {num_accounts} account"
    header += "s" if num_accounts > 1 else ""
    click.echo(f"{header}\n")

    for index in range(num_accounts):
        account = starknet_accounts[index]
        output_dict = {
            "Alias": account.alias,
            "Public key": account.public_key,
            "Class hash": to_hex(account.class_hash),
        }
        if not account.deployments:
            output_dict["Contract address (not deployed)"] = account.address

        else:
            for deployment in account.deployments:
                key = f"Contract address ({deployment.network_name})"
                output_dict[key] = deployment.contract_address

        output_keys = add_padding_to_strings(list(output_dict.keys()))
        output_dict = {k: output_dict[k.rstrip()] for k in output_keys}

        for k, v in output_dict.items():
            click.echo(f"{k} - {v}")

        if index < num_accounts - 1:
            click.echo()


def _validate_network(parse, network: str) -> str:
    # Temporarily connect to extract network name and validate network.
    live_nets = list(NETWORKS.keys())
    options = [*live_nets, LOCAL_NETWORK_NAME]

    if network.startswith("starknet:"):
        network = network.replace("starknet:", "")

    if network not in options:
        options_str = ",".join(options)
        raise click.BadOptionUsage("--network", f"Network '{network}' not one of '{options_str}'.")

    return f"starknet:{network}"


@accounts.command(name="import")
@ape_cli_context()
@click.argument("alias")
@network_option(required=True)
@address_option()
@class_hash_option()
@click.option("--salt", help="The contract address salt used when deploying the contract.")
def _import(cli_ctx, alias, network, address, class_hash, salt):
    """Add an existing, deployed account"""

    container = _get_container(cli_ctx)
    if alias in container.aliases:
        # Since the alias exists, try to import as a new deployment.

        existing_account = container.load(alias)
        if not isinstance(existing_account, StarknetKeyfileAccount):
            cli_ctx.abort(
                "Unable to use default ape-starknet CLI to add "
                "deployments to accounts from other plugins. "
                f"Try using plugin with account type {type(existing_account).__name__}."
            )

        else:
            # Assume calculated address if not given anything.
            address = address or existing_account.address
            if not salt and address == existing_account.address:
                # It must have used the same salt since the addresses are the same.
                salt = existing_account.salt

            networks_str = ", ".join(network)
            click.echo(f"Importing existing account on network(s) '{networks_str}'.")
            for network_name in network:
                existing_account.add_deployment(network_name, address, salt, leave_unlocked=False)

    elif address and not class_hash:
        cli_ctx.abort("--class-hash is required when importing an account for the first time.")

    elif address:
        # Account is being imported for the first time.
        private_key = click.prompt(f"Enter private key for '{alias}'", hide_input=True).strip()

        if private_key.isnumeric():
            # NOTE: Check base 10 before 16 because is_hex("123") is True.
            private_key = int(private_key)
        elif is_hex(private_key):
            private_key = int(private_key, 16)
        else:
            cli_ctx.abort("Invalid private key. Expecting numeric value.")

        if private_key < 1 or private_key >= EC_ORDER:
            cli_ctx.abort("Private key not in range [1, EC_ORDER).")

        deployments = [
            StarknetAccountDeployment(contract_address=address, network_name=n, salt=salt)
            for n in network
        ]
        container.import_account(alias, class_hash, private_key, salt=salt, deployments=deployments)

    else:
        cli_ctx.abort("--address required when importing and account for the first time.")

    cli_ctx.logger.success(f"Imported account '{alias}'.")


@accounts.command()
@ape_cli_context()
@existing_alias_argument(account_type=BaseStarknetAccount)
@network_option()
@address_option()
def delete(cli_ctx, alias, network, address):
    """Delete an existing account deployment"""
    container = _get_container(cli_ctx)
    account = container.load(alias)
    deployments_before = len(account.deployments)
    container.delete_account(alias, networks=network, address=address, leave_unlocked=False)

    if account.address_int not in container:
        cli_ctx.logger.success(f"Account '{alias}' deleted.")
    elif (len(account.deployments) < deployments_before) and network:
        choices_str = ",".join([x.replace("starknet:", "") for x in network])
        cli_ctx.logger.success(f"Account '{alias}' deployments on network '{choices_str}' deleted.")


@accounts.command()
@ape_cli_context()
@existing_alias_argument(account_type=StarknetKeyfileAccount)
def change_password(cli_ctx, alias):
    """Change the password of an existing account"""
    account = cli_ctx.account_manager.load(alias)
    account.change_password(leave_unlocked=False)
    cli_ctx.logger.success(f"Password has been changed for account '{alias}'")



import json
import shutil
from contextlib import contextmanager
from pathlib import Path
from tempfile import mkdtemp
from typing import cast

import ape
import pytest
from ape.api.networks import LOCAL_NETWORK_NAME
from ape.contracts import ContractInstance
from ethpm_types import ContractType
from starkware.cairo.lang.compiler.test_utils import short_string_to_felt

from ape_starknet import tokens as _tokens
from ape_starknet.accounts import StarknetAccountContainer, StarknetKeyfileAccount
from ape_starknet.utils import OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH, PLUGIN_NAME

# NOTE: Ensure that we don't use local paths for these
ape.config.DATA_FOLDER = Path(mkdtemp()).resolve()
ape.config.PROJECT_FOLDER = Path(mkdtemp()).resolve()

_HERE = Path(__file__).parent
projects_directory = Path(__file__).parent / "projects"
project_names = [p.stem for p in projects_directory.iterdir() if p.is_dir()]
ALIAS = "__TEST_ALIAS__"
SECOND_ALIAS = "__TEST_ALIAS_2__"
EXISTING_KEY_FILE_ALIAS = f"{ALIAS}existing_key_file"
EXISTING_EPHEMERAL_ALIAS = f"{ALIAS}existing_ephemeral"
PASSWORD = "123"
PUBLIC_KEY = 367323783092256132793135877206902311054243182689252847590820585364953599456

# Pre-checksummed
CONTRACT_ADDRESS = "0x06b7243aA4eDBe5bD629c6712B3ac9639B160480a7730a31483F7b387463A183"

START_BALANCE = "1000000 ETH"

# Purposely pick a number larest enough to test Uint256 logic
TOKEN_INITIAL_SUPPLY = 2 * 2**128

ETH_CONTRACT_TYPE = ContractType.parse_obj(
    {
        "abi": [
            {
                "anonymous": False,
                "inputs": [
                    {"indexed": False, "name": "prevNum", "type": "uint256"},
                    {"indexed": True, "name": "newNum", "type": "uint256"},
                ],
                "name": "NumberChange",
                "type": "event",
            },
            {"inputs": [], "stateMutability": "nonpayable", "type": "constructor"},
            {
                "inputs": [{"name": "num", "type": "uint256"}],
                "name": "setNumber",
                "outputs": [],
                "stateMutability": "nonpayable",
                "type": "function",
            },
            {
                "inputs": [],
                "name": "owner",
                "outputs": [{"name": "", "type": "address"}],
                "stateMutability": "view",
                "type": "function",
            },
            {
                "inputs": [],
                "name": "myNumber",
                "outputs": [{"name": "", "type": "uint256"}],
                "stateMutability": "view",
                "type": "function",
            },
            {
                "inputs": [],
                "name": "prevNumber",
                "outputs": [{"name": "", "type": "uint256"}],
                "stateMutability": "view",
                "type": "function",
            },
        ],
        "contractName": "EthContract",
        "deploymentBytecode": {
            "bytecode": "0x3360005561012261001c6300000000396101226000016300000000f3600436101561000d57610117565b60003560e01c3461011d57633fb5c1cb81186100d057600054331461008957600b6040527f21617574686f72697a65640000000000000000000000000000000000000000006060526040506040518060600181600003601f1636823750506308c379a06000526020602052601f19601f6040510116604401601cfd5b60056004351461011d576001546002556004356001556004357f2295d5ec33e3af0d43cc4b73aa3cd7d784150fe365cbdb4b4fd338220e4f135760025460405260206040a2005b638da5cb5b81186100e75760005460405260206040f35b6323fd0e4081186100fe5760015460405260206040f35b634825cf6f81186101155760025460405260206040f35b505b60006000fd5b600080fd"  # noqa: E501
        },
        "devdoc": {},
        "runtimeBytecode": {
            "bytecode": "0x600436101561000d57610117565b60003560e01c3461011d57633fb5c1cb81186100d057600054331461008957600b6040527f21617574686f72697a65640000000000000000000000000000000000000000006060526040506040518060600181600003601f1636823750506308c379a06000526020602052601f19601f6040510116604401601cfd5b60056004351461011d576001546002556004356001556004357f2295d5ec33e3af0d43cc4b73aa3cd7d784150fe365cbdb4b4fd338220e4f135760025460405260206040a2005b638da5cb5b81186100e75760005460405260206040f35b6323fd0e4081186100fe5760015460405260206040f35b634825cf6f81186101155760025460405260206040f35b505b60006000fd5b600080fd"  # noqa: E501
        },
        "sourceId": "EthContract.vy",
        "userdoc": {},
    }
)


@pytest.fixture(scope="session")
def existing_key_file_alias():
    return EXISTING_KEY_FILE_ALIAS


@pytest.fixture(scope="session")
def contract_address():
    return CONTRACT_ADDRESS


@pytest.fixture(scope="session")
def password():
    return PASSWORD


@pytest.fixture(scope="session")
def public_key():
    return PUBLIC_KEY


@pytest.fixture(scope="session")
def token_initial_supply():
    return TOKEN_INITIAL_SUPPLY


@pytest.fixture(scope="session")
def project_path():
    return projects_directory / "project"


@pytest.fixture(scope="session")
def token_project_path():
    return projects_directory / "token"


@pytest.fixture(scope="session")
def proxy_project_path():
    return projects_directory / "proxy"


@pytest.fixture(scope="session")
def data_folder():
    return Path(__file__).parent / "data"


@pytest.fixture(scope="session")
def config():
    return ape.config


@pytest.fixture(scope="session")
def accounts():
    return ape.accounts


@pytest.fixture
def networks():
    return ape.networks


@pytest.fixture(scope="session")
def convert():
    return ape.convert


@pytest.fixture(scope="session")
def chain():
    return ape.chain


@pytest.fixture(scope="session")
def project(contracts):
    with contracts.use_project() as proj:
        yield proj


@pytest.fixture(scope="session")
def use_local_starknet():
    choice = f"{PLUGIN_NAME}:{LOCAL_NETWORK_NAME}:{PLUGIN_NAME}"
    return ape.networks.parse_network_choice(choice)


@pytest.fixture(scope="session")
def use_local_ethereum():
    return ape.networks.parse_network_choice(f"ethereum:{LOCAL_NETWORK_NAME}")


@pytest.fixture(scope="session", autouse=True)
def connected(use_local_starknet):
    with use_local_starknet:
        yield


@pytest.fixture(scope="session")
def tokens():
    return _tokens


@pytest.fixture
def provider(chain):
    return chain.provider


@pytest.fixture(scope="session")
def explorer(use_local_starknet):
    with use_local_starknet as provider:
        return provider


@pytest.fixture(scope="session")
def eth_contract_type():
    return ETH_CONTRACT_TYPE


@pytest.fixture(autouse=True, scope="session")
def clean_projects():
    def clean():
        for project in projects_directory.iterdir():
            if not project.is_dir() or project.name.startswith("."):
                continue

            cache_dir = project / ".build"
            if not cache_dir.is_dir():
                continue

            shutil.rmtree(cache_dir)

    clean()
    yield
    clean()


@pytest.fixture(scope="session")
def project_switcher(config, project_path, token_project_path, proxy_project_path):
    projects = {"project": project_path, "token": token_project_path, "proxy": proxy_project_path}

    class ProjectSwitcher:
        @contextmanager
        def use(self, name: str):
            with config.using_project(projects[name]) as project:
                yield project

    return ProjectSwitcher()


@pytest.fixture(scope="session")
def contracts(
    clean_projects,
    project_switcher,
    account,
    token_initial_supply,
    use_local_starknet,
):
    class ContractsDeployer:
        """
        A solution to more lazily deploy session-scoped contracts.
        This is helpful to speed up tests when not using them.
        """

        @property
        def my_contract(self) -> ContractInstance:
            with self.use_project() as project:
                if project.MyContract.deployments:
                    return project.MyContract.deployments[-1]

                # Declare + Deploy contract only once.
                account.declare(project.MyContract)
                contract = project.MyContract.deploy(sender=account)
                contract.initialize(sender=account)
                return contract

        @property
        def token(self):
            with self.use_project(name="token") as project:
                if project.TestToken.deployments:
                    return project.TestToken.deployments[-1]

                account.declare(project.TestToken)
                name = short_string_to_felt("TestToken")
                symbol = short_string_to_felt("TEST")
                contract = project.TestToken.deploy(
                    name, symbol, 18, token_initial_supply, int(account.address, 16), sender=account
                )
                _tokens.add_token("test_token", LOCAL_NETWORK_NAME, contract.address)
                return contract

        @property
        def user_token(self):
            with self.use_project(name="token") as project:
                if project.UseToken.deployments:
                    return project.UseToken.deployments[-1]

                account.declare(project.UseToken)
                return project.UseToken.deploy(sender=account)

        @property
        def proxy(self):
            with self.use_project(name="proxy") as project:
                if project.Proxy.deployments:
                    return project.Proxy.deployments[-1]

                account.declare(project.Proxy)
                contract = project.Proxy.deploy(self.token.address, sender=account)
                _tokens.add_token("proxy_token", LOCAL_NETWORK_NAME, contract.address)
                return contract

        @contextmanager
        def use_project(self, name: str = "project"):
            with use_local_starknet:
                with project_switcher.use(name) as proj:
                    yield proj

    return ContractsDeployer()


@pytest.fixture(scope="session")
def contract(contracts):
    return contracts.my_contract


@pytest.fixture(scope="session")
def token_contract(contracts):
    return contracts.token


@pytest.fixture(scope="session")
def token_user_contract(contracts):
    return contracts.user_token


@pytest.fixture
def initial_balance(contract, account):
    return contract.get_balance(account.address)


@pytest.fixture(scope="session")
def account_container(accounts):
    return cast(StarknetAccountContainer, accounts.containers[PLUGIN_NAME])


@pytest.fixture(scope="session")
def account(account_container, use_local_starknet):
    with use_local_starknet:
        return account_container.test_accounts[0]


@pytest.fixture(scope="session")
def second_account(account_container, use_local_starknet):
    with use_local_starknet:
        return account_container.test_accounts[1]


@pytest.fixture(scope="session")
def eth_account(accounts):
    return accounts.test_accounts[0]


@pytest.fixture(scope="session")
def ephemeral_account(account_container):
    return create_account(account_container)


def create_account(container):
    return container.create_account(ALIAS)


@pytest.fixture(scope="session")
def starknet():
    return ape.networks.starknet


@pytest.fixture(scope="session")
def testnet(starknet):
    return starknet.testnet


@pytest.fixture
def in_starknet_testnet(testnet):
    with testnet.use_provider("starknet"):
        yield


@pytest.fixture(scope="session")
def key_file_account_data():
    return {
        "ape-starknet": {
            "public_key": 367323783092256132793135877206902311054243182689252847590820585364953599456,  # noqa: E501
            "class_hash": 3146761231686369291210245479075933162526514193311043598334639064078158562617,  # noqa: E501
            "salt": 2383401134194309956567840095871463496630094772532838902518655252027359569593,
            "deployments": [
                {"network_name": "testnet", "contract_address": CONTRACT_ADDRESS},
            ],
        },
        "address": "93920847a9ab3c731562461bb7d0fb95a39df9f9",
        "crypto": {
            "cipher": "aes-128-ctr",
            "cipherparams": {"iv": "b279075a04c21ed621e466974c64e6e3"},
            "ciphertext": "e3fb18fbd034df62963ac776c04fec9a58a71273cbbd236fa6f22f98a7563e5a",
            "kdf": "scrypt",
            "kdfparams": {
                "dklen": 32,
                "n": 262144,
                "r": 1,
                "p": 8,
                "salt": "810c59cb704d6012ce2002012cbdf735",
            },
            "mac": "f903f6d42efd883e3b402f68c7681857080a889a363e78e7694e143b5a9a9a65",
        },
        "id": "2b50ad6b-d136-4981-b09c-08c006e2ae56",
        "version": 3,
    }


@pytest.fixture(scope="session")
def ephemeral_account_data():
    return {
        "private_key": 509219664670742235607272813021130138373595301613956902800973975925797957544,
        "public_key": 2068822281043178075870469557539081791152169138879468456959920393634230618024,
        "class_hash": OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH,
    }


@pytest.fixture
def key_file_account(config, key_file_account_data):
    temp_accounts_dir = Path(config.DATA_FOLDER) / "starknet"
    temp_accounts_dir.mkdir(exist_ok=True, parents=True)
    test_key_file_path = temp_accounts_dir / f"{EXISTING_KEY_FILE_ALIAS}.json"

    if test_key_file_path.is_file():
        test_key_file_path.unlink()

    test_key_file_path.write_text(json.dumps(key_file_account_data))

    account = StarknetKeyfileAccount(key_file_path=test_key_file_path)
    account.unlock(passphrase="123")

    yield account

    if test_key_file_path.is_file():
        test_key_file_path.unlink()


@pytest.fixture(scope="session")
def traces_testnet_243810(data_folder):
    content = (data_folder / "traces-testnet-block-243810.json").read_text()
    return json.loads(content)["traces"]


@pytest.fixture(scope="session")
def traces_testnet_243810_results(data_folder):
    content = (data_folder / "traces-testnet-block-243810_results.json").read_text()
    return json.loads(content)



import os
from dataclasses import asdict
from typing import Dict, Iterator, List, Optional, Union, cast
from urllib.error import HTTPError
from urllib.parse import urlparse
from urllib.request import urlopen

from ape.api import BlockAPI, ProviderAPI, ReceiptAPI, SubprocessProvider, TransactionAPI
from ape.api.networks import LOCAL_NETWORK_NAME
from ape.exceptions import ProviderNotConnectedError, TransactionError
from ape.logging import logger
from ape.types import AddressType, BlockID, ContractLog, LogFilter, RawAddress
from ape.utils import DEFAULT_NUMBER_OF_TEST_ACCOUNTS, cached_property, raises_not_implemented
from requests import Session
from starknet_py.net.client_errors import ContractNotFoundError
from starknet_py.net.client_models import (
    BlockSingleTransactionTrace,
    ContractCode,
    SentTransactionResponse,
    StarknetBlock,
)
from starknet_py.net.gateway_client import GatewayClient
from starknet_py.net.models import parse_address
from starkware.starkware_utils.error_handling import StarkErrorCode

from ape_starknet.config import DEFAULT_PORT, StarknetConfig
from ape_starknet.exceptions import StarknetProviderError
from ape_starknet.transactions import (
    AccountTransaction,
    DeclareTransaction,
    DeployAccountTransaction,
    InvokeFunctionTransaction,
    StarknetTransaction,
)
from ape_starknet.utils import (
    ALPHA_MAINNET_WL_DEPLOY_TOKEN_KEY,
    DEFAULT_ACCOUNT_SEED,
    DEVNET_ACCOUNT_START_BALANCE,
    EXECUTE_METHOD_NAME,
    EXECUTE_SELECTOR,
    PLUGIN_NAME,
    get_chain_id,
    get_class_hash,
    get_dict_from_tx_info,
    handle_client_error,
    handle_client_errors,
    run_until_complete,
    to_checksum_address,
    to_int,
)
from ape_starknet.utils.basemodel import StarknetBase


class DevnetClient:
    def __init__(self, host_address: str):
        self.session = Session()
        self.host_address = host_address

    @cached_property
    def predeployed_accounts(self) -> List[Dict]:
        return self._get("predeployed_accounts")

    def increase_time(self, amount: int):
        return self._post("increase_time", json={"time": amount})

    def create_block(self):
        return self._post("create_block")

    def set_time(self, amount: int):
        return self._post("set_time", json={"time": amount})

    def mint(self, address: str, amount: int):
        return self._post("mint", json={"address": address, "amount": amount})

    def _get(self, uri: str, **kwargs):
        return self._request("get", uri, **kwargs)

    def _post(self, uri: str, **kwargs):
        return self._request("post", uri, **kwargs)

    def _request(self, method: str, uri: str, **kwargs):
        response = self.session.request(method.upper(), url=f"{self.host_address}/{uri}", **kwargs)
        response.raise_for_status()
        return response.json()


class StarknetProvider(ProviderAPI, StarknetBase):
    """
    A Starknet provider.
    """

    # Gets set when 'connect()' is called.
    client: Optional[GatewayClient] = None
    cached_code: Dict[int, ContractCode] = {}

    # Use int address to be free from duplicate keys.
    local_nonce_cache: Dict[int, int] = {}

    @property
    def connected_client(self) -> GatewayClient:
        if not self.client:
            raise ProviderNotConnectedError()

        return self.client

    @property
    def is_connected(self) -> bool:
        was_successful = False
        try:
            urlopen(self.uri)
            was_successful = True
        except HTTPError as err:
            was_successful = err.code == 404  # Task failed successfully
        except Exception:
            was_successful = False

        if was_successful and self.client is None:
            self.client = self._create_client()

        return was_successful

    @property
    def starknet_client(self) -> GatewayClient:
        if not self.is_connected:
            raise StarknetProviderError("Provider is not connected to Starknet.")

        return self.client

    @cached_property
    def plugin_config(self) -> StarknetConfig:
        return self.config_manager.get_config(PLUGIN_NAME) or StarknetConfig()  # type: ignore

    @cached_property
    def uri(self) -> str:
        network_config = self.plugin_config.provider.dict().get(self.network.name)
        if not network_config:
            raise StarknetProviderError(f"Unknown network '{self.network.name}'.")

        return network_config.get("uri") or f"http://127.0.0.1:{DEFAULT_PORT}"

    def connect(self):
        self.client = self._create_client()

    def disconnect(self):
        self.client = None
        super().disconnect()

    def update_settings(self, new_settings: dict):
        pass

    @property
    def chain_id(self) -> int:
        return get_chain_id(self.network.name).value

    @handle_client_errors
    def get_balance(self, address: AddressType) -> int:
        return self.tokens.get_balance(address)

    @handle_client_errors
    def get_code(self, address: str) -> List[int]:
        # NOTE: Always return truthy value for code so that Ape core works properly
        return self.get_code_and_abi(address).bytecode or [ord(c) for c in "PROXY"]

    @handle_client_errors
    def get_abi(self, address: str) -> List[Dict]:
        return self.get_code_and_abi(address).abi

    @handle_client_errors
    def get_nonce(self, address: RawAddress) -> int:
        if self.network.name != LOCAL_NETWORK_NAME:
            return self.connected_client.get_contract_nonce_sync(address)

        address_int = int(address) if isinstance(address, int) else int(address, 16)
        if address_int not in self.local_nonce_cache:
            nonce = self.connected_client.get_contract_nonce_sync(address)
            self.local_nonce_cache[address_int] = nonce

        return self.local_nonce_cache[address_int]

    @handle_client_errors
    def estimate_gas_cost(self, txn: StarknetTransaction) -> int:
        if isinstance(txn, InvokeFunctionTransaction):
            if not txn.is_prepared:
                txn = cast(InvokeFunctionTransaction, self.prepare_transaction(txn))

            if txn.method_abi.name != EXECUTE_METHOD_NAME:
                txn = txn.as_execute()

        if not txn.signature:
            # Signature is required to estimate gas, unfortunately.
            # the transaction is typically signed by this point, but not always.
            txn.signature = self.account_manager[txn.receiver].sign_transaction(txn)

        starknet_object = txn.as_starknet_object()
        estimated_fee = self.connected_client.estimate_fee_sync(starknet_object)
        return estimated_fee.overall_fee

    @property
    def gas_price(self) -> int:
        """
        **NOTE**: Currently, the gas price is fixed to always be 100 gwei.
        """

        return self.conversion_manager.convert("100 gwei", int)

    @handle_client_errors
    def get_block(self, block_id: BlockID) -> BlockAPI:
        block = self._get_block(block_id)
        return self.starknet.decode_block(block)

    def _get_block(self, block_id: BlockID) -> StarknetBlock:
        if isinstance(block_id, bytes):
            block_id = to_int(block_id)

        if isinstance(block_id, (int, str)) and len(str(block_id)) >= 72:
            kwarg = "block_hash"
        elif block_id in ("pending", "latest"):
            kwarg = "block_number"
        elif isinstance(block_id, int):
            kwarg = "block_number"
            if block_id < 0:
                latest_block_number = self.get_block("latest").number
                block_id_int = latest_block_number + block_id + 1
                if block_id_int < 0:
                    raise StarknetProviderError(f"Block with number '{block_id_int}' not found.")

                block_id = block_id_int

        else:
            raise StarknetProviderError(f"Unsupported BlockID type '{type(block_id)}'.")

        return self.starknet_client.get_block_sync(**{kwarg: block_id})

    @handle_client_errors
    def send_call(self, txn: TransactionAPI) -> bytes:
        if not isinstance(txn, InvokeFunctionTransaction):
            type_str = f"{txn.type!r}" if isinstance(txn.type, bytes) else str(txn.type)
            raise StarknetProviderError(
                f"Transaction must be from an invocation. Received type {type_str}."
            )

        starknet_obj = txn._as_call()
        return self.connected_client.call_contract_sync(starknet_obj)

    @handle_client_errors
    def _get_traces(self, block_number: int) -> List[BlockSingleTransactionTrace]:
        block_traces = self.starknet_client.get_block_traces_sync(block_number=block_number)
        return block_traces.traces

    @handle_client_errors
    def _get_single_trace(
        self, block_number: int, txn_hash: int
    ) -> Optional[BlockSingleTransactionTrace]:
        traces = self._get_traces(block_number)
        return next((trace for trace in traces if trace.transaction_hash == txn_hash), None)

    @handle_client_errors
    def get_receipt(
        self, txn_hash: str, transaction: Optional[TransactionAPI] = None
    ) -> ReceiptAPI:
        self.starknet_client.wait_for_tx_sync(txn_hash)
        txn_info, receipt = run_until_complete(
            self.starknet_client.get_transaction(txn_hash),
            self.starknet_client.get_transaction_receipt(tx_hash=txn_hash),
        )
        data = {**asdict(receipt), **get_dict_from_tx_info(txn_info)}
        was_deploy = False
        # Handle __execute__ overhead. User only cares for target ABI.
        if data.get("entry_point_selector") == EXECUTE_SELECTOR:
            num_calls = data["calldata"][0]
            if num_calls != 1:
                logger.warning("Multi-call not yet supported. Only parsing first receipt.")

            data["sender"] = data["contract_address"]
            data["contract_address"] = self.starknet.decode_address(data["calldata"][1])
            data["entry_point_selector"] = data["calldata"][2]
            stop_index = data["calldata"][3] + 1
            data["calldata"] = data["calldata"][4:stop_index]

        if not transaction:
            transaction = self.starknet.create_transaction(**data)

        if isinstance(transaction, DeployAccountTransaction):
            # The DeployAccountReceipt expects first-hand access to the contract address.
            data["contract_address"] = transaction.contract_address
        elif isinstance(transaction, InvokeFunctionTransaction):
            # Manually remove `contract_address`, as it always confused Pydantic because
            # of TransactionAPI's `contract_address` attribute, which is required but a different
            # concept than this (for deploys, it is the new contract address but for invoke, it's
            # the receiver).
            data["receiver"] = data.pop("contract_address")
            original_tx = transaction.original_transaction
            was_deploy = (
                original_tx is not None
                and original_tx.method_abi.name == "deployContract"
                and int(original_tx.receiver, 16) == int(self.universal_deployer.address, 16)
            )

        receipt = self.starknet.decode_receipt(
            {"provider": self, "transaction": transaction, **data}
        )

        if was_deploy:
            event_abi = self.universal_deployer.contract_type.events["ContractDeployed"]
            logs = list(receipt.decode_logs(event_abi))
            if not logs:
                raise StarknetProviderError("Failed to create contract.")

            # Set the receipt `contract_address` to the newly deployed contract's address,
            # mimicing how it works in Ethereum.
            receipt.contract_address = self.starknet.decode_address(logs[-1]["address"])

        return receipt

    def get_transactions_by_block(self, block_id: BlockID) -> Iterator[TransactionAPI]:
        block = self._get_block(block_id)
        for txn_info in block.transactions:
            txn_dict = get_dict_from_tx_info(txn_info)
            yield self.starknet.create_transaction(**txn_dict)

    @handle_client_errors
    def send_transaction(self, txn: TransactionAPI, token: Optional[str] = None) -> ReceiptAPI:
        response = self._send_transaction(txn, token=token)
        if response.code != StarkErrorCode.TRANSACTION_RECEIVED.name:
            raise TransactionError(message="Transaction not received.")

        receipt = self.get_receipt(response.transaction_hash, transaction=txn)
        if not isinstance(txn, AccountTransaction):
            return receipt

        sender_address = None
        if isinstance(txn, InvokeFunctionTransaction):
            # The real sender is the account contract address.
            sender_address = txn.receiver
        elif isinstance(txn, DeclareTransaction):
            sender_address = txn.sender
        elif isinstance(txn, DeployAccountTransaction):
            sender_address = to_checksum_address(txn.contract_address)

        if sender_address:
            address_int = int(sender_address, 16)
            if address_int in self.local_nonce_cache:
                self.local_nonce_cache[address_int] += 1
            else:
                # Trigger setting nonce via call
                _ = self.provider.get_nonce(address_int)

            total_cost = receipt.total_fees_paid + receipt.value
            if total_cost:
                self.tokens.update_cache(address_int, -total_cost)

        return receipt

    def _send_transaction(
        self, txn: TransactionAPI, token: Optional[str] = None
    ) -> SentTransactionResponse:
        if not isinstance(txn, StarknetTransaction):
            raise StarknetProviderError(
                "Unable to send non-Starknet transaction using a Starknet provider "
                f"(received type '{type(txn)}')."
            )
        elif (
            isinstance(txn, InvokeFunctionTransaction)
            and txn.method_abi.name != EXECUTE_METHOD_NAME
        ):
            raise StarknetProviderError(
                f"Can only send invoke transaction to an account {EXECUTE_METHOD_NAME} method."
            )

        env_token = os.environ.get(ALPHA_MAINNET_WL_DEPLOY_TOKEN_KEY)
        token = token or getattr(txn, "token", env_token)
        starknet_txn = txn.as_starknet_object()
        result = self.starknet_client.send_transaction_sync(starknet_txn, token=token)
        return result

    @raises_not_implemented
    def get_contract_logs(  # type: ignore[empty-body]
        self, log_filter: LogFilter
    ) -> Iterator[ContractLog]:
        pass

    def prepare_transaction(self, txn: TransactionAPI) -> TransactionAPI:
        txn.chain_id = self.provider.chain_id

        # All preparation happens on the account side.
        if isinstance(txn, AccountTransaction) and not txn.is_prepared and txn.sender:
            account = self.account_container[txn.sender]
            return account.prepare_transaction(txn)

        return txn

    def get_virtual_machine_error(self, exception: Exception):
        return handle_client_error(exception)

    def get_code_and_abi(self, address: Union[str, AddressType, int]) -> ContractCode:
        address_int = parse_address(address)

        # Cache code for faster look-up
        if address_int not in self.cached_code:
            try:
                code_and_abi = self.starknet_client.get_code_sync(address_int)
            except ContractNotFoundError:
                # Nothing deployed to this address.
                code_and_abi = ContractCode(bytecode=[], abi=[])

            self.cached_code[address_int] = code_and_abi

        return self.cached_code[address_int]

    def get_class_hash(self, address: AddressType) -> int:
        code = self.get_code_and_abi(address)
        return get_class_hash(code)

    def _create_client(self) -> GatewayClient:
        network = self.uri if self.network.name == LOCAL_NETWORK_NAME else self.network.name
        return GatewayClient(network)


class StarknetDevnetProvider(SubprocessProvider, StarknetProvider):
    """
    A subprocess provider for the starknet-devnet process.
    """

    @property
    def process_name(self) -> str:
        return "starknet-devnet"

    @cached_property
    def devnet_client(self) -> DevnetClient:
        return DevnetClient(self.uri)

    def connect(self):
        if self.network.name == LOCAL_NETWORK_NAME:
            # Behave like a 'SubprocessProvider'
            if not self.is_connected:
                super().connect()

            self.start()

            # Ensure test accounts get cached so they remain when switching providers.
            _ = self.account_container.test_accounts

        self.client = self._create_client()

    def build_command(self) -> List[str]:
        parts = urlparse(self.uri)
        return [
            self.process_name,
            "--host",
            str(parts.hostname),
            "--port",
            str(parts.port),
            "--accounts",
            str(DEFAULT_NUMBER_OF_TEST_ACCOUNTS),
            "--seed",
            str(DEFAULT_ACCOUNT_SEED),
            "--initial-balance",
            str(DEVNET_ACCOUNT_START_BALANCE),
        ]

    def set_timestamp(self, new_timestamp: int):
        if self.devnet_client is None:
            raise StarknetProviderError("Must be connected to starknet-devnet to use this feature.")

        result = self.devnet_client.set_time(new_timestamp)
        if "next_block_timestamp" not in result:
            raise StarknetProviderError(result)

    def mine(self, num_blocks: int = 1):
        for index in range(num_blocks):
            self.devnet_client.create_block()

    def set_balance(self, account: AddressType, amount: Union[int, float, str, bytes]):
        """
        Mint the difference between the account's current balance and the desired balance
        and give it to the account.
        """

        # Convert account
        if not isinstance(account, str):
            account = self.conversion_manager.convert(account, AddressType)

        # Convert amount
        if isinstance(amount, str) and " " in amount:
            # Convert values like "10 ETH"
            amount = self.conversion_manager.convert(amount, int)
        elif not isinstance(amount, int):
            amount = to_int(amount)

        amount = int(amount)  # for mypy
        current_balance = self.get_balance(account)
        difference = amount - current_balance
        if difference < 0:
            raise StarknetProviderError(
                f"Unable to set balance to {amount}, "
                f"as it is less than the current balance of {current_balance}."
            )

        self.devnet_client.mint(account, difference)
        self.tokens.update_cache(account, difference)


__all__ = ["StarknetProvider", "StarknetDevnetProvider"]



from functools import reduce
from typing import Any, List, Optional

from ape.api import TransactionAPI
from pydantic import BaseModel
from starkware.crypto.signature.signature import pedersen_hash

from ape_starknet.utils import to_int


def _prepare_message(message: Any) -> List[int]:
    if not isinstance(message, (list, tuple)):
        message = [message]

    converted: List[int] = []
    for value in message:
        if isinstance(value, StarknetSignableMessage):
            converted.extend(_prepare_message(value.message))
        elif isinstance(value, (list, tuple)):
            converted.extend(_prepare_message(value))
        elif isinstance(value, TransactionAPI):
            converted.append(to_int(value.txn_hash))
        else:
            converted.append(to_int(value))

    return converted


class StarknetSignableMessage(BaseModel):
    message: Optional[Any]

    @property
    def message_ints(self) -> List[int]:
        return _prepare_message(self.message)

    @property
    def hash(self) -> int:
        return reduce(lambda x, y: pedersen_hash(y, x), self.message_ints, 0)

    def __str__(self) -> str:
        return str(self.message)

    def __repr__(self) -> str:
        return f"<message={self.message}>"



import asyncio
import re
from asyncio import gather
from dataclasses import asdict
from json import JSONDecodeError, loads
from typing import Any, Dict, List, Optional, Union, cast

from ape.api.networks import LOCAL_NETWORK_NAME
from ape.contracts import ContractEvent
from ape.exceptions import (
    ApeException,
    ContractError,
    ContractLogicError,
    OutOfGasError,
    SignatureError,
)
from ape.logging import logger
from ape.types import AddressType, RawAddress
from eth_typing import HexAddress, HexStr
from eth_utils import add_0x_prefix, is_0x_prefixed, is_hex, is_text, remove_0x_prefix, to_hex
from eth_utils import to_int as eth_to_int
from ethpm_types import ContractType
from ethpm_types.abi import EventABI, MethodABI
from hexbytes import HexBytes
from starknet_py.net import KeyPair
from starknet_py.net.client_errors import ClientError
from starknet_py.net.client_models import (
    BlockSingleTransactionTrace,
    DeclareTransaction,
    DeployAccountTransaction,
    InvokeTransaction,
    Transaction,
)
from starknet_py.net.models import TransactionType
from starknet_py.net.models.address import parse_address
from starknet_py.transaction_exceptions import TransactionRejectedError
from starkware.cairo.bootloaders.compute_fact import keccak_ints
from starkware.crypto.signature.signature import get_random_private_key as get_random_pkey
from starkware.starknet.core.os.class_hash import compute_class_hash
from starkware.starknet.definitions.general_config import StarknetChainId
from starkware.starknet.public.abi import get_selector_from_name
from starkware.starknet.services.api.contract_class import ContractClass
from starkware.starknet.third_party.open_zeppelin.starknet_contracts import account_contract

from ape_starknet.exceptions import StarknetProviderError
from ape_starknet.utils.basemodel import create_contract_class

PLUGIN_NAME = "starknet"
NETWORKS = {
    # chain_id, network_id
    "mainnet": (StarknetChainId.MAINNET.value, StarknetChainId.MAINNET.value),
    "testnet": (StarknetChainId.TESTNET.value, StarknetChainId.TESTNET.value),
    "testnet2": (StarknetChainId.TESTNET2.value, StarknetChainId.TESTNET.value),
}
_HEX_ADDRESS_REG_EXP = re.compile("(0x)?[0-9a-f]*", re.IGNORECASE | re.ASCII)
"""Same as from eth-utils except not limited length."""
ALPHA_MAINNET_WL_DEPLOY_TOKEN_KEY = "ALPHA_MAINNET_WL_DEPLOY_TOKEN"
EXECUTE_METHOD_NAME = "__execute__"
EXECUTE_SELECTOR = get_selector_from_name(EXECUTE_METHOD_NAME)
DEFAULT_ACCOUNT_SEED = 2147483647  # Prime
ContractEventABI = Union[List[Union[EventABI, ContractEvent]], Union[EventABI, ContractEvent]]
MAX_FEE = int(1e20)
_DECLARE_ERROR_PATTERN = re.compile(r"Class with hash (0x[0-9a-fA-F]+) is not declared")
STARKNET_FEE_TOKEN_SYMBOL = "ETH"
_CLIENT_FAILED_PREFIX_PATTERN = re.compile(r"Client failed( with code \d+)?: (.*)")


def convert_contract_class_to_contract_type(
    name: str, source_id: str, contract_class: ContractClass
):
    return ContractType.parse_obj(
        {
            "contractName": name,
            "sourceId": source_id,
            "deploymentBytecode": {"bytecode": contract_class.serialize().hex()},
            "runtimeBytecode": {},
            "abi": contract_class.abi,
        }
    )


OPEN_ZEPPELIN_ACCOUNT_SOURCE_ID = "openzeppelin/account/Account.cairo"
OPEN_ZEPPELIN_ACCOUNT_CONTRACT_TYPE = convert_contract_class_to_contract_type(
    "Account", OPEN_ZEPPELIN_ACCOUNT_SOURCE_ID, account_contract
)
OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH = compute_class_hash(account_contract)
EXECUTE_ABI = OPEN_ZEPPELIN_ACCOUNT_CONTRACT_TYPE.mutable_methods[EXECUTE_METHOD_NAME]

# Taken from https://github.com/argentlabs/argent-x/blob/develop/packages/extension/src/background/wallet.ts  # noqa: E501
ARGENTX_ACCOUNT_CLASS_HASH = int(
    "0x025ec026985a3bf9d0cc1fe17326b245dfdc3ff89b8fde106542a3ea56c5a918", 16
)
ARGENTX_ACCOUNT_SOURCE_ID = "account/ArgentAccount.cairo"
DEVNET_ACCOUNT_START_BALANCE = 1000000000000000000000


def get_chain_id(network_id: Union[str, int]) -> StarknetChainId:
    if isinstance(network_id, int):
        return StarknetChainId(network_id)

    elif network_id == LOCAL_NETWORK_NAME:
        return StarknetChainId.TESTNET  # Use TESTNET chain ID for local network

    elif network_id not in NETWORKS:
        raise StarknetProviderError(f"Unknown network '{network_id}'.")

    return StarknetChainId(NETWORKS[network_id][0])


def to_checksum_address(address: RawAddress) -> AddressType:
    if is_checksum_address(address):
        return cast(AddressType, address)

    return _to_checksum_address(address)


def _to_checksum_address(address: RawAddress) -> AddressType:
    if isinstance(address, bytes):
        address = HexBytes(address).hex()

    address_int = parse_address(address)
    address_str = pad_hex_str(HexBytes(address_int).hex().lower())
    chars = [c for c in remove_0x_prefix(HexStr(address_str))]
    hashed = [b for b in HexBytes(keccak_ints([address_int]))]

    for i in range(0, len(chars), 2):
        try:
            if hashed[i >> 1] >> 4 >= 8:
                chars[i] = chars[i].upper()
            if (hashed[i >> 1] & 0x0F) >= 8:
                chars[i + 1] = chars[i + 1].upper()
        except IndexError:
            continue

    rejoined_address_str = add_0x_prefix(HexStr("".join(chars)))
    return AddressType(HexAddress(HexStr(rejoined_address_str)))


def is_hex_address(value: Any) -> bool:
    return _HEX_ADDRESS_REG_EXP.fullmatch(value) is not None if is_text(value) else False


def is_checksum_address(value: Any) -> bool:
    if not is_text(value):
        return False

    if not is_hex_address(value):
        return False

    return value == _to_checksum_address(value)


def extract_trace_data(trace: BlockSingleTransactionTrace) -> Dict[str, Any]:
    if not trace:
        return {}

    trace_data = trace.function_invocation

    # Keep the most relevant `result`: given the account implementation, `result`
    # may contain an additional number prepend to the data to expose the total
    # number of items. For a method returning a 3-items array like `[1, 2, 3]`,
    # in such scenario `results` would be `[0x4, 0x3, 0x1, 0x2, 0x3]` (the prepend
    # number: 4, the array length: 3, then array items: 1, 2, and 3).
    # As there is no known way to guess when to remove such a number, we prefer to "scan"
    # trace internals to select the most appropriate result. For instance, when `result`
    # contains the additional value, we just need to use the "internal call" `result`
    # that will contain the exact value the method returned.
    invocation_result = trace_data["result"]
    internal_calls = trace_data["internal_calls"]
    trace_data["result"] = (
        internal_calls[-1]["result"] if internal_calls else invocation_result
    ) or invocation_result
    return trace_data


def handle_client_errors(f):
    def func(*args, **kwargs):
        try:
            result = f(*args, **kwargs)
            if isinstance(result, dict) and result.get("error"):
                message = result["error"].get("message") or "Transaction failed"
                raise StarknetProviderError(message)

            return result

        except Exception as err:
            new_err = handle_client_error(err)
            if new_err:
                raise new_err from err

            raise  # Original error

    return func


def handle_client_error(err: Exception) -> Optional[Exception]:
    if isinstance(err, ApeException) or not isinstance(
        err, (ClientError, TransactionRejectedError)
    ):
        return err

    err_msg = err.message
    if "Actual fee exceeded max fee" in err_msg:
        return OutOfGasError()

    maybe_contract_logic_error = False
    if "Error message:" in err_msg:
        err_msg = err_msg.split("Error message:")[-1].splitlines()[0].strip().split("\\n")[0]
        maybe_contract_logic_error = True

    elif "Error at pc=" in err_msg:
        err_msg = err_msg.strip().replace("\\n", " ")
        err_msg = _try_extract_message_from_json(err_msg)
        maybe_contract_logic_error = True

    if "INVALID_SIGNATURE_LENGTH" in err_msg:
        return SignatureError("Invalid signature length.")

    elif "UNINITIALIZED_CONTRACT" in err_msg and "is not deployed" in err_msg:
        address = err_msg.split("is not deployed")[0].strip().split(" ")[-1]
        if is_hex(address):
            return ContractError(f"Contract at address '{address}' not deployed.")

    elif "Signature" in err_msg and "is invalid, with respect to the public key" in err_msg:
        message = "Invalid signature"
        parts = err_msg.split("public key ")
        if len(parts) == 2:
            key_str = parts[-1].split(" ")[0].rstrip(",")
            if key_str.isnumeric():
                key = to_hex(int(key_str))
                message = f"{message} with respect to public key {key}"

        return SignatureError(f"{message}.")

    declare_error_search = _DECLARE_ERROR_PATTERN.search(err_msg)
    if declare_error_search:
        address = declare_error_search.groups()[0]
        return StarknetProviderError(f"Contract with address '{address}' not declared.")

    elif maybe_contract_logic_error:
        return ContractLogicError(revert_message=err_msg)

    err_msg = err_msg.strip().rstrip(".")
    match = _CLIENT_FAILED_PREFIX_PATTERN.match(err_msg)
    if match:
        groups = match.groups()
        if groups:
            trimmed = groups[-1]
            if trimmed:
                err_msg = trimmed

    if err_msg.endswith("\n."):
        err_msg = "\n.".join(err_msg.split("\n.")[:-1])

    err_msg = _try_extract_message_from_json(err_msg)

    if not err_msg.endswith("."):
        err_msg = f"{err_msg}."

    return StarknetProviderError(err_msg)


def _try_extract_message_from_json(value: str) -> str:
    try:
        msg_dict = loads(value)
    except JSONDecodeError:
        return value

    if "message" in msg_dict:
        return msg_dict["message"]

    return value


def get_dict_from_tx_info(txn_info: Transaction) -> Dict:
    txn_dict = {**asdict(txn_info)}
    if isinstance(txn_info, InvokeTransaction):
        txn_dict["type"] = TransactionType.INVOKE_FUNCTION
    elif isinstance(txn_info, DeclareTransaction):
        txn_dict["type"] = TransactionType.DECLARE
    elif isinstance(txn_info, DeployAccountTransaction):
        txn_dict["type"] = TransactionType.DEPLOY_ACCOUNT

    return txn_dict


def get_method_abi_from_selector(
    selector: Union[str, int], contract_type: ContractType
) -> MethodABI:
    # TODO: Properly integrate with ethpm-types

    if isinstance(selector, str):
        selector = int(selector, 16)

    for abi in contract_type.mutable_methods:
        selector_to_check = get_selector_from_name(abi.name)

        if selector == selector_to_check:
            return abi

    raise ContractError(f"Method '{selector}' not found in '{contract_type.name}'.")


def get_random_private_key() -> str:
    private_key = HexBytes(get_random_pkey()).hex()
    return pad_hex_str(private_key)


def pad_hex_str(value: str, to_length: int = 64) -> str:
    val = value.replace("0x", "")
    actual_len = len(val)
    padding = "0" * (to_length - actual_len)
    return f"0x{padding}{val}"


def run_until_complete(*coroutine):
    coroutines = list(coroutine)
    if len(coroutines) > 1:
        task = gather(*coroutine, return_exceptions=True)
    else:
        task = coroutines[0]

    loop = asyncio.get_event_loop()
    result = loop.run_until_complete(task)
    return result


def to_int(val: Any) -> int:
    if isinstance(val, int):
        return val

    elif isinstance(val, str) and is_0x_prefixed(val):
        return eth_to_int(hexstr=val)

    elif isinstance(val, str) and val.isnumeric():
        return int(val)

    elif isinstance(val, str):
        return eth_to_int(val.encode())

    elif hasattr(val, "address"):
        return to_int(val.address)

    return eth_to_int(val)


def get_class_hash(code: Union[str, HexBytes]):
    contract_class = create_contract_class(code)
    return compute_class_hash(contract_class)


def create_keypair(private_key: Union[str, int]) -> KeyPair:
    # Validate private key.
    if isinstance(private_key, str) and is_0x_prefixed(private_key):
        private_key = to_int(private_key.strip("'\""))
    elif isinstance(private_key, str):
        private_key = to_int(private_key)

    return KeyPair.from_private_key(private_key)


def get_account_constructor_calldata(key_pair: KeyPair, class_hash: int) -> List[Any]:
    # Use known ctor data
    if class_hash == OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH:
        return [key_pair.public_key]
    elif class_hash == ARGENTX_ACCOUNT_CLASS_HASH:
        return []
    else:
        logger.warning(f"Constructor calldata for account with class '{class_hash}' not known.")
        return []



import pytest
from ape.exceptions import (
    ApeException,
    ContractError,
    ContractLogicError,
    OutOfGasError,
    SignatureError,
)
from hexbytes import HexBytes
from starknet_py.net.client_errors import ClientError, ContractNotFoundError
from starknet_py.net.client_models import BlockSingleTransactionTrace
from starknet_py.transaction_exceptions import TransactionRejectedError

from ape_starknet.exceptions import StarknetProviderError
from ape_starknet.utils import (
    extract_trace_data,
    get_random_private_key,
    handle_client_error,
    is_checksum_address,
    to_checksum_address,
    to_int,
)

TEST_ADDRESS = 852629284295565304522725188288313475155928521990240236795402253191102056828


@pytest.mark.parametrize("iteration", range(10))
def test_get_random_private_key(iteration):
    pkey = get_random_private_key()
    assert len(pkey) == 66
    pkey_int = int(pkey, 16)
    pkey_back_to_str = HexBytes(pkey_int).hex()
    assert pkey_back_to_str.replace("0x", "") in pkey


def test_is_checksum_address(account):
    assert is_checksum_address(account.address)


def test_to_checksum_address(account):
    # Value from Starknet.js result
    expected = "0x02F4F57e5948B113Bf3D807B9ABB900EfC689b5b7a8500EBb79F670B3e08AE24"
    actual = to_checksum_address(expected.lower())
    assert actual == expected


@pytest.mark.parametrize(
    "exception, expected",
    [
        (ApeException("Foo!"), ApeException("Foo!")),
        (
            ClientError(
                message=(
                    '{"message":"Error at pc=0:91:\nGot an exception '
                    "while executing a hint.\nCairo traceback (most r"
                    "ecent call last):\nUnknown location (pc=0:739)\n"
                    "Unknown location (pc=0:682)\nUnknown location (p"
                    "c=0:358)\nUnknown location (pc=0:400)\nUnknown l"
                    "ocation (pc=0:423)\n\nError in the called contra"
                    "ct (0x123):\nError at pc=0:41:\nGot an exception"
                    " while executing a hint.\nCairo traceback (most "
                    "recent call last):\nUnknown location (pc=0:7453)"
                    "\nUnknown location (pc=0:7437)\nUnknown location"
                    " (pc=0:4491)\nError message: Strings: exceeding "
                    "max felt string length (31)\nUnknown location (p"
                    "c=0:3219)\nUnknown location (pc=0:47)\n\nTraceba"
                    'ck (most recent call last):\n  File "<hint6>", l'
                    "ine 3, in <module>\nAssertionError: a = 36185027"
                    "886661312136973227830950701056231072153315966999"
                    '20561358720 is out of range.","status_code":500}'
                ),
            ),
            ContractLogicError(revert_message="Strings: exceeding max felt string length (31)"),
        ),
        (
            ClientError(
                message=(
                    '{"message":"An InvokeFunction transaction '
                    '(version != 0) must have a nonce.","status_code":500}\n'
                )
            ),
            StarknetProviderError(
                "An InvokeFunction transaction (version != 0) must have a nonce."
            ),
        ),
        (
            ContractNotFoundError(
                address=TEST_ADDRESS,
                block_hash="pending",
            ),
            StarknetProviderError(
                f"No contract with address {TEST_ADDRESS} found for block with block_hash: pending."
            ),
        ),
        (
            TransactionRejectedError(message="Actual fee exceeded max fee.\n999800000000000 > 1"),
            OutOfGasError(),
        ),
        (
            TransactionRejectedError(
                message="Error at pc=0:330: An ASSERT_EQ instruction failed: 0 != 1."
            ),
            ContractLogicError(
                revert_message="Error at pc=0:330: An ASSERT_EQ instruction failed: 0 != 1."
            ),
        ),
        (ValueError("Foo!"), ValueError("Foo!")),
        (
            ClientError(
                "Client failed with code 500: "
                '{"code": "StarknetErrorCode.TRANSACTION_FAILED", '
                '"message": "Error at pc=0:166:\\nSignature '
                "(3043690392760996823501689233597619912806801642077770433986032017604167509951, "
                "140932343948720055880187056217327835615175238698469767628052057966817353503), "
                "is invalid, with respect to the public key "
                "6835981319243216192375995835136133760276036836, "
                "and the message hash "
                "2506859578009568926921040670359334110481328447009243612404313172189339494201."
                "\\nCairo traceback (most recent call last):\\nUnknown "
                "location (pc=0:413)\\nUnknown location (pc=0:400)"
                '\\nUnknown location (pc=0:320)"}.'
            ),
            SignatureError(
                "Invalid signature with respect to public key "
                "0x13289378ec83a20385758c7ec489853213cfce4."
            ),
        ),
        (
            ClientError(
                "Client failed with code 500: "
                '{{"code":"StarknetErrorCode.UNINITIALIZED_CONTRACT",'
                f'"message":"Requested contract address {TEST_ADDRESS} is not deployed."}}\n.'
            ),
            ContractError(f"Contract at address '{TEST_ADDRESS}' not deployed."),
        ),
    ],
)
def test_handle_client_error(exception, expected):
    error = handle_client_error(exception)
    assert str(error) == str(expected)


def test_extract_trace_data(traces_testnet_243810, traces_testnet_243810_results):
    for trace in traces_testnet_243810:
        trace_object = BlockSingleTransactionTrace(**trace)
        trace_data = extract_trace_data(trace_object)
        assert isinstance(trace_data, dict)

        expected_result = traces_testnet_243810_results[trace_object.transaction_hash]
        assert trace_data["result"] == expected_result


@pytest.mark.parametrize(
    "val", (1952805748, "1952805748", b"test", "test", HexBytes(1952805748), "0x74657374")
)
def test_to_int(val):
    assert to_int(val) == 1952805748



from typing import Iterator, Optional, Union, cast

from ape.api import ExplorerAPI, ReceiptAPI
from ape.types import AddressType
from ape.utils import raises_not_implemented
from ethpm_types import ContractType, HexBytes

from ape_starknet.accounts import BaseStarknetAccount
from ape_starknet.utils import EXECUTE_METHOD_NAME, pad_hex_str
from ape_starknet.utils.basemodel import StarknetBase


class StarknetExplorer(ExplorerAPI, StarknetBase):
    BASE_URIS = {
        "testnet": "https://goerli.voyager.online",
        "testnet2": "https://goerli-2.voyager.online",
        "mainnet": "https://voyager.online",
    }

    @property
    def base_uri(self) -> str:
        network_name = self.provider.network.name
        return self.BASE_URIS.get(network_name, "")

    def get_address_url(self, address: AddressType) -> str:
        base_uri = self.base_uri
        return f"{base_uri}/contract/{address}" if base_uri else ""

    def get_transaction_url(self, transaction_hash: str) -> str:
        base_uri = self.base_uri
        return f"{base_uri}/tx/{transaction_hash}" if base_uri else ""

    def get_contract_type(self, address: Union[AddressType, int]) -> Optional[ContractType]:
        if self.tokens.is_token(address):
            return self.tokens.contract_type

        elif address in self.account_container:
            starknet_account = cast(BaseStarknetAccount, self.account_container[address])
            return starknet_account.contract_type

        return self.get_contract_type_from_provider(address)

    def get_contract_type_from_provider(self, address: Union[int, AddressType]):
        code_and_abi = self.provider.get_code_and_abi(address)

        # Convert list of ints to bytes
        bytecode_str_list = [HexBytes(x).hex() for x in code_and_abi.bytecode]
        if not bytecode_str_list:
            return None

        longest_str = min(len(max(bytecode_str_list, key=len)) - 2, 0)
        code_parts = [
            pad_hex_str(x, to_length=longest_str).replace("0x", "") for x in bytecode_str_list
        ]

        contract_type_dict = {
            "abi": code_and_abi.abi,
            "deploymentBytecode": {"bytecode": f"0x{''.join(code_parts)}"},
        }

        if EXECUTE_METHOD_NAME in [a["name"] for a in code_and_abi.abi]:
            contract_type_dict["contractName"] = "Account"

        return ContractType(**contract_type_dict)

    @raises_not_implemented
    def get_account_transactions(  # type: ignore[empty-body]
        self, address: AddressType
    ) -> Iterator[ReceiptAPI]:
        # TODO
        pass

    @raises_not_implemented
    def publish_contract(self, address: AddressType):
        # TODO
        pass



from typing import Literal

import pytest
from ape.types import AddressType
from eth_typing import HexAddress, HexStr
from ethpm_types.abi import MethodABI
from hexbytes import HexBytes
from starkware.starknet.public.abi import get_selector_from_name

INT_ADDRESS = 269168490721327376227480949634158339134330130662514218728945045982440529971
STR_ADDRESS = "0x0098580E36aB1485C66F0Dc95c2c923E734B7aF44d04Dd2B5B9d0809Aa672033"
HEXBYTES_ADDRESS = HexBytes(STR_ADDRESS)
EVENT_NAME = "balance_increased"


class CustomABI(MethodABI):
    name: str = "Custom"
    type: Literal["function"] = "function"


@pytest.fixture(scope="module")
def raw_logs():
    return [
        {
            "data": ["4321"],
            "from_address": "0x14acf3b7e92f97adee4d5359a7de3d673582f0ce03d33879cdbdbf03ec7fa5d",
            "keys": [get_selector_from_name(EVENT_NAME)],
            "transaction_hash": "0x7ccac756aafe0df416ee4f4ca74b42cdd7399ebae5aa31d92132dfbee445370",
            "block_number": 5,
            "block_hash": "0x7032c67f8741c6ce547175c2101d5ccdf468ca6ea0642bafccea04080726a06",
        }
    ]


@pytest.fixture
def event_abi(contract):
    return contract.balance_increased.abi


@pytest.mark.parametrize("value", (INT_ADDRESS, STR_ADDRESS, HEXBYTES_ADDRESS))
def test_encode_and_decode_address(value, starknet):
    decoded_address = starknet.decode_address(value)
    expected = AddressType(HexAddress(HexStr(STR_ADDRESS)))
    assert decoded_address == expected

    # The values should _always_ encode back to the INT_ADDRESS.
    re_encoded_address = starknet.encode_address(decoded_address)
    assert re_encoded_address == INT_ADDRESS


def test_decode_logs(starknet, event_abi, raw_logs):
    actual = list(starknet.decode_logs(raw_logs, event_abi))
    assert len(actual) == 1
    assert actual[0].amount == "4321"



from copy import deepcopy
from dataclasses import asdict
from typing import Any, Dict, Iterator, List, Optional

from ape.api import ReceiptAPI, TransactionAPI
from ape.exceptions import APINotImplementedError, TransactionError
from ape.types import AddressType, ContractLog
from ape.utils import abstractmethod, cached_property, raises_not_implemented
from ethpm_types import ContractType, HexBytes
from ethpm_types.abi import EventABI, MethodABI
from pydantic import Field, validator
from starknet_py.net.client_models import Call, Event, TransactionStatus
from starknet_py.net.models.transaction import (
    Declare,
    DeployAccount,
    InvokeFunction,
    Transaction,
    TransactionType,
)
from starkware.starknet.core.os.contract_address.contract_address import (
    calculate_contract_address_from_hash,
)
from starkware.starknet.core.os.transaction_hash.transaction_hash import (
    TransactionHashPrefix,
    calculate_declare_transaction_hash,
    calculate_deploy_account_transaction_hash,
    calculate_transaction_hash_common,
)
from starkware.starknet.definitions import constants
from starkware.starknet.definitions.fields import ContractAddressSalt
from starkware.starknet.public.abi import get_selector_from_name
from starkware.starknet.services.api.contract_class import ContractClass

from ape_starknet.exceptions import ContractTypeNotFoundError
from ape_starknet.utils import (
    EXECUTE_ABI,
    EXECUTE_METHOD_NAME,
    OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH,
    OPEN_ZEPPELIN_ACCOUNT_CONTRACT_TYPE,
    ContractEventABI,
    extract_trace_data,
    to_checksum_address,
    to_int,
)
from ape_starknet.utils.basemodel import StarknetBase


class StarknetTransaction(TransactionAPI, StarknetBase):
    """
    A base transaction class for all Starknet transactions.
    """

    version: int = constants.TRANSACTION_VERSION

    class Config:
        use_enum_values = True

    def __str__(self) -> str:
        data = self.dict()
        params = "\n  ".join(
            f"{k}: {v}" for k, v in data.items() if k not in ("data", "method_abi", "signature")
        )
        return f"{self.__class__.__name__}:\n  {params}"

    def serialize_transaction(self) -> dict:  # type: ignore
        return self.dict()

    @abstractmethod
    def as_starknet_object(self) -> Transaction:
        """
        Convert :class:`~ape.api.providers.TransactionAPI` to its Starknet
        transaction equivalent so it can be accepted by the core Starknet OS
        framework.
        """

    @property
    def total_transfer_value(self) -> int:
        return self.value + (self.max_fee or 0)


class AccountTransaction(StarknetTransaction):
    """
    Transactions that must go through an account contract.
    (Invoke and Declare).
    """

    max_fee: int = Field(0)
    nonce: Optional[int] = None
    is_prepared: bool = Field(False, exclude=True)

    @validator("max_fee", pre=True, allow_reuse=True)
    def validate_max_fee(cls, value):
        if isinstance(value, str):
            return int(value, 16)

        return value or 0

    @property
    def starknet_signature(self) -> List[int]:
        if self.signature:
            return [to_int(self.signature.r), to_int(self.signature.s)]

        return []


class DeclareTransaction(AccountTransaction):
    sender: AddressType = Field(alias="sender_address")
    type: TransactionType = TransactionType.DECLARE

    @validator("sender", pre=True, allow_reuse=True)
    def validate_sender(cls, value):
        return to_checksum_address(value)

    @cached_property
    def starknet_contract(self) -> ContractClass:
        return ContractClass.deserialize(self.data)

    @property
    def txn_hash(self) -> HexBytes:
        return calculate_declare_transaction_hash(
            self.starknet_contract,
            self.provider.chain_id,
            self.max_fee,
            to_int(self.sender),
            self.version,
            self.nonce,
        )

    def as_starknet_object(self) -> Declare:
        return Declare(
            contract_class=self.starknet_contract,
            max_fee=self.max_fee,
            nonce=self.nonce,
            sender_address=self.starknet.encode_address(self.sender),
            signature=self.starknet_signature,
            version=self.version,
        )


class InvokeFunctionTransaction(AccountTransaction):
    method_abi: MethodABI
    sender: Optional[AddressType] = None
    type: TransactionType = TransactionType.INVOKE_FUNCTION

    contract_address: Optional[AddressType]
    """
    Gets set when calling `deployContract` on a UDC contract.
    """

    # Gets set when calling `as_execute()` and is intended to be the
    # transaction before transforming to an `__execute__()`transaction.
    original_transaction: Optional["InvokeFunctionTransaction"] = Field(
        None, exclude=True, repr=False
    )
    """
    The original transaction before transforming to an account ``__execute__()``
    invoke transaction. Gets set in the
    ``:meth:`~ape_starknet.transactions.InvokeFunctionTransaction.as_execute`` method.
    """

    data: List[Any] = Field([], alias="calldata")  # type: ignore
    receiver: AddressType

    def __str__(self) -> str:
        # Show original transaction in str so it is easier to tell what it is.
        # (as opposed to the __execute__ call on an account txn that makes a real call).
        if self.method_abi.name == EXECUTE_METHOD_NAME and self.original_transaction is not None:
            return str(self.original_transaction).replace("max_fee: 0", f"max_fee: {self.max_fee}")

        return super().__str__()

    @validator("receiver", pre=True, allow_reuse=True)
    def validate_receiver(cls, value):
        return to_checksum_address(value)

    @validator("max_fee", pre=True, allow_reuse=True)
    def validate_max_fee(cls, value):
        return to_int(value)

    @property
    def receiver_int(self) -> int:
        return self.starknet.encode_address(self.receiver)

    @property
    def contract_type(self) -> ContractType:
        contract_type = self.get_contract_type(self.receiver)
        if not contract_type:
            raise ContractTypeNotFoundError(self.receiver)

        return contract_type

    @property
    def entry_point_selector(self) -> int:
        return get_selector_from_name(self.method_abi.name)

    @property
    def txn_hash(self) -> HexBytes:
        hash_int = calculate_transaction_hash_common(
            additional_data=[],
            calldata=self.data,
            chain_id=self.chain_id,
            contract_address=self.receiver_int,
            entry_point_selector=self.entry_point_selector,
            max_fee=self.max_fee or 0,
            tx_hash_prefix=TransactionHashPrefix.INVOKE,
            version=self.version,
        )
        return HexBytes(hash_int)

    def as_starknet_object(self) -> InvokeFunction:
        return InvokeFunction(
            calldata=self.data,
            contract_address=self.receiver_int,
            max_fee=self.max_fee or 0,
            nonce=self.nonce,
            signature=self.starknet_signature,
            version=self.version,
        )

    def _as_call(self) -> InvokeFunction:
        receiver_int = self.starknet.encode_address(self.receiver)
        return Call(to_addr=receiver_int, selector=self.entry_point_selector, calldata=self.data)

    def as_execute(self) -> "InvokeFunctionTransaction":
        """
        Convert this transaction to an account ``__execute__`` transaction.
        """

        new_tx = deepcopy(self)
        stark_tx = new_tx.as_starknet_object()
        account_call = {
            "to": stark_tx.contract_address,
            "selector": new_tx.entry_point_selector,
            "data_offset": 0,
            "data_len": len(stark_tx.calldata),
        }
        full_abi = OPEN_ZEPPELIN_ACCOUNT_CONTRACT_TYPE.abi
        entire_call_data = [[account_call], stark_tx.calldata]
        new_tx.data = new_tx.starknet.encode_calldata(full_abi, EXECUTE_ABI, entire_call_data)

        if new_tx.sender:
            new_tx.receiver = new_tx.sender

        new_tx.sender = None
        new_tx.method_abi = EXECUTE_ABI
        new_tx.original_transaction = self
        return new_tx


class DeployAccountTransaction(AccountTransaction):
    salt: int = Field(alias="contract_address_salt")
    class_hash: int = OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH
    constructor_calldata: List[Any]
    nonce: int = 0
    deployer_contract_address: int = 0
    type: TransactionType = TransactionType.DEPLOY_ACCOUNT

    @validator("salt", pre=True, allow_reuse=True)
    def validate_salt(cls, value):
        return value or ContractAddressSalt.get_random_value()

    @property
    def contract_address(self) -> int:
        return calculate_contract_address_from_hash(
            class_hash=self.class_hash,
            constructor_calldata=self.constructor_calldata,
            deployer_address=self.deployer_contract_address,
            salt=self.salt,
        )

    @property
    def txn_hash(self) -> HexBytes:
        return calculate_deploy_account_transaction_hash(
            version=self.version,
            contract_address=self.contract_address,
            class_hash=self.class_hash,
            constructor_calldata=self.constructor_calldata,
            max_fee=self.max_fee,
            nonce=self.nonce,
            salt=self.salt,
            chain_id=self.chain_id,
        )

    def as_starknet_object(self) -> DeployAccount:
        return DeployAccount(
            contract_address_salt=self.salt,
            class_hash=self.class_hash,
            constructor_calldata=self.constructor_calldata,
            nonce=self.nonce,
            signature=self.starknet_signature,
            max_fee=self.max_fee,
            version=self.version,
        )


class StarknetReceipt(ReceiptAPI, StarknetBase):
    """
    An object represented a confirmed transaction in Starknet.
    """

    status: TransactionStatus
    # TODO: Figure out why None sometimes
    block_hash: Optional[str] = None

    """Aliased"""
    txn_hash: str = Field(alias="hash")
    gas_used: int = Field(alias="actual_fee")

    @property
    def return_value(self) -> Any:
        raise APINotImplementedError("'return_value' can only be accessed on InvokeTransactions")

    @validator("block_hash", pre=True, allow_reuse=True)
    def validate_block_hash(cls, value):
        return HexBytes(value).hex() if value else value

    @validator("txn_hash", pre=True, allow_reuse=True)
    def validate_transaction_hash(cls, value):
        if isinstance(value, int):
            return HexBytes(value).hex()

    @property
    def ran_out_of_gas(self) -> bool:
        return False  # Overidden by Invoke receipts

    @property
    def total_fees_paid(self) -> int:
        return 0  # Overidden by Account receipts

    @raises_not_implemented
    def decode_logs(  # type: ignore[empty-body]
        self, abi: Optional[ContractEventABI] = None
    ) -> Iterator[ContractLog]:
        # Overriden in InvocationReceipt
        pass


class AccountTransactionReceipt(StarknetReceipt):
    @property
    def total_fees_paid(self) -> int:
        return self.gas_used  # gas_used is a misleading name.


class DeployAccountReceipt(AccountTransactionReceipt):
    contract_address: AddressType
    status: TransactionStatus

    @validator("contract_address", pre=True, allow_reuse=True)
    def validate_contract_address(cls, value):
        if isinstance(value, str) and not value.startswith("0x"):
            return to_checksum_address(int(value))

        return to_checksum_address(value)


class InvokeFunctionReceipt(AccountTransactionReceipt):
    logs: List[dict] = Field(alias="events")

    @validator("logs", pre=True, allow_reuse=True)
    def validate_logs(cls, value):
        if value and isinstance(value[0], Event):
            value = [asdict(event) for event in value]

        return value

    @property
    def ran_out_of_gas(self) -> bool:
        return self.gas_used >= (self.max_fee or 0)

    @cached_property
    def trace(self) -> Dict:  # type: ignore
        trace = self.provider._get_single_trace(self.block_number, to_int(self.txn_hash))
        return extract_trace_data(trace) if trace else {}

    @property
    def returndata(self):
        return self.trace.get("result", [])

    @cached_property
    def return_value(self) -> Any:
        txn = self.transaction
        if not isinstance(txn, InvokeFunctionTransaction):
            return None  # Should never get here.

        if txn.original_transaction:
            txn = txn.original_transaction

        method_abi = txn.method_abi
        return self.starknet.decode_returndata(method_abi, self.returndata)

    def decode_logs(
        self,
        abi: Optional[ContractEventABI] = None,
    ) -> Iterator[ContractLog]:

        log_data_items: List[Dict] = []
        for log in self.logs:
            log_data = {
                **log,
                "block_hash": self.block_hash,
                "transaction_hash": self.txn_hash,
                "block_number": self.block_number,
            }
            log_data_items.append(log_data)

        if abi is not None:
            if not isinstance(abi, (list, tuple)):
                abi = [abi]

            event_abis: List[EventABI] = [a.abi if not isinstance(a, EventABI) else a for a in abi]
            yield from self.starknet.decode_logs(log_data_items, *event_abis)

        else:
            # If ABI is not provided, decode all events
            address_map = {
                x["from_address"]: self.starknet.decode_address(x["from_address"])
                for x in log_data_items
            }
            contract_types = self.chain_manager.contracts.get_multiple(address_map.values())
            # address → selector → abi
            selectors = {
                address: {get_selector_from_name(e.name): e for e in contract.events}
                for address, contract in contract_types.items()
            }
            for log in log_data_items:
                contract_address = address_map[log["from_address"]]
                if contract_address not in selectors:
                    continue

                for event_key in log.get("keys", []):
                    event_abi = selectors[contract_address][event_key]
                    yield from self.starknet.decode_logs([log], event_abi)


class ContractDeclaration(AccountTransactionReceipt):
    """
    The result of declaring a contract type in Starknet.
    """

    class_hash: int

    @validator("class_hash", pre=True)
    def validate_class_hash(cls, value):
        if isinstance(value, str):
            return int(value, 16)
        elif isinstance(value, bytes):
            return to_int(value)
        else:
            return value

    @cached_property
    def contract_type(
        self,
    ) -> ContractType:
        contract_type = self.get_local_contract_type(self.class_hash)
        if not contract_type:
            raise TransactionError(message="Contract type declaration was not successful.")

        return contract_type


__all__ = [
    "ContractDeclaration",
    "InvokeFunctionTransaction",
    "StarknetReceipt",
    "StarknetTransaction",
]



import pytest
from ape.api.networks import LOCAL_NETWORK_NAME

from .conftest import ApeStarknetCliRunner


@pytest.fixture(scope="session")
def console_runner(ape_cli):
    return ApeStarknetCliRunner(
        ape_cli, ["console", "--network", f"ethereum:{LOCAL_NETWORK_NAME}:test"]
    )


def test_console_accounts_object(ape_cli, console_runner, key_file_account, networks):
    # NOTE: This console connects to Eth-Tester and makes sure we can still _read_
    # starknet accounts.
    output = console_runner.invoke(
        input=["accounts", f"accounts['{key_file_account.address}']", "exit"]
    )
    assert key_file_account.address in output, output



import pytest
from ape.api.networks import LOCAL_NETWORK_NAME
from click.testing import CliRunner
from hexbytes import HexBytes

from ape_starknet.accounts import DEVNET_CONTRACT_SALT, StarknetAccountDeployment
from ape_starknet.utils import OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH, is_hex_address


@pytest.fixture
def give_input():
    def fn(msg):
        return CliRunner().isolation(input=msg)

    return fn


@pytest.fixture
def devnet_keyfile_account(account_container, account, password, give_input):
    deployments = [
        StarknetAccountDeployment(
            contract_address=account.address, network_name="testnet", salt=DEVNET_CONTRACT_SALT
        )
    ]

    with give_input(f"{password}\n{password}\n"):
        return account_container.import_account(
            "__DEV_AS_KEYFILE_ACCOUNT__",
            OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH,
            account.private_key,
            deployments=deployments,
            salt=DEVNET_CONTRACT_SALT,
        )


@pytest.fixture
def txn(devnet_keyfile_account):
    return devnet_keyfile_account.get_deploy_account_txn()


def test_public_key(key_file_account, public_key):
    actual = key_file_account.public_key
    assert int(actual, 16) == public_key


def test_address(account):
    assert is_hex_address(account.address)


def test_account_container_contains(account, second_account, key_file_account, account_container):
    assert account.address in account_container
    assert second_account.address in account_container
    assert key_file_account.address in account_container


CASE_NAME_TO_LAMBDA = {
    "address_str": lambda a, _: a.address,
    "address_int": lambda a, e: e.encode_address(a.address),
    "public_key_str": lambda a, _: a.public_key,
    "public_key_int": lambda a, e: e.encode_address(a.public_key),
}


@pytest.mark.parametrize("case", [x for x in CASE_NAME_TO_LAMBDA.keys()])
def test_access_account_by_str_address(account, account_container, starknet, case):
    get_address = CASE_NAME_TO_LAMBDA[case]
    address = get_address(account, starknet)
    assert account_container[address] == account
    assert address in account_container


def test_balance(account, ephemeral_account, tokens):
    balance = account.balance
    assert isinstance(balance, int)
    assert balance > 0

    balance = ephemeral_account.balance
    assert isinstance(balance, int)
    assert balance > 0

    # Clear caches and make sure still works (uses RPC)
    del tokens.balance_cache[account.address_int]
    del tokens.balance_cache[ephemeral_account.address_int]

    balance = account.balance
    assert isinstance(balance, int)
    assert balance > 0

    balance = ephemeral_account.balance
    assert isinstance(balance, int)
    assert balance > 0


def test_can_access_devnet_accounts(account, second_account, chain):
    assert chain.contracts[account.address] == account.contract_type
    assert chain.contracts[second_account.address] == second_account.contract_type


def test_import_with_passphrase(account_container, account, give_input):
    alias = "__TEST_IMPORT_WITH_PASSPHRASE__"
    deployment = StarknetAccountDeployment(
        contract_address=account.address, network_name=LOCAL_NETWORK_NAME, salt=DEVNET_CONTRACT_SALT
    )

    with give_input("p@55W0rd\np@55W0rd\n"):
        account_container.import_account(
            alias,
            OPEN_ZEPPELIN_ACCOUNT_CLASS_HASH,
            account.private_key,
            deployments=[deployment],
        )
    new_account = account_container.load(alias)
    assert new_account.address == account.address


def test_transfer(account, second_account):
    initial_account_balance = account.balance
    initial_second_account_balance = second_account.balance
    receipt = account.transfer(second_account, 10)
    total_cost = receipt.total_fees_paid + 10
    assert second_account.balance == initial_second_account_balance + 10
    assert account.balance == initial_account_balance - total_cost


@pytest.mark.parametrize("msg", (5, [5, 6], b"5", HexBytes(122)))
def test_sign_message(msg, key_file_account, password):
    key_file_account.set_autosign(True, passphrase=password)
    assert key_file_account.sign_message(msg)
    key_file_account.set_autosign(False)


def test_unlock_with_passphrase_and_sign_message(
    in_starknet_testnet, give_input, devnet_keyfile_account, password
):
    devnet_keyfile_account.unlock(passphrase=password)
    with give_input("y\n"):
        devnet_keyfile_account.sign_message([1, 2, 3])


def test_unlock_from_prompt_and_sign_message(
    in_starknet_testnet, give_input, devnet_keyfile_account, password
):
    with give_input(f"{password}\n"):
        devnet_keyfile_account.unlock()

    with give_input("y\n"):
        devnet_keyfile_account.sign_message([1, 2, 3])


def test_unlock_with_passphrase_and_sign_transaction(
    in_starknet_testnet, give_input, devnet_keyfile_account, password, txn
):
    devnet_keyfile_account.unlock(passphrase=password)
    with give_input("y\n"):
        devnet_keyfile_account.sign_transaction(txn)


def test_unlock_from_prompt_and_sign_transaction(
    in_starknet_testnet, give_input, devnet_keyfile_account, password, txn
):
    with give_input(f"{password}\n"):
        devnet_keyfile_account.unlock()

    with give_input("y\n"):
        devnet_keyfile_account.sign_transaction(txn)


def test_set_autosign(in_starknet_testnet, give_input, devnet_keyfile_account, password, txn):
    with give_input(f"{password}\n"):
        devnet_keyfile_account.set_autosign(True)

    devnet_keyfile_account.sign_transaction(txn)

    # Disable and verify we have to sign again
    devnet_keyfile_account.set_autosign(False)
    with give_input(f"y\n{password}\n"):
        devnet_keyfile_account.sign_transaction(txn)


def test_set_autosign_and_provide_passphrase(
    in_starknet_testnet, give_input, devnet_keyfile_account, password, txn
):
    devnet_keyfile_account.set_autosign(True, passphrase=password)
    devnet_keyfile_account.sign_transaction(txn)

    # Disable and verify we have to sign again
    devnet_keyfile_account.set_autosign(False)
    with give_input(f"y\n{password}\n"):
        devnet_keyfile_account.sign_transaction(txn)


def test_accounts_devnet_accounts_are_still_available_on_ethereum(account_container, networks):
    assert account_container.test_accounts

    with networks.ethereum.local.use_provider("test"):
        # Accounts are still found even though devnet is not the active provider.
        assert account_container.test_accounts

    assert account_container.test_accounts



from functools import cached_property
from typing import TYPE_CHECKING, Optional, Union, cast

from ape.types import AddressType
from ape.utils import ManagerAccessMixin
from eth_utils import is_0x_prefixed
from ethpm_types import ContractType
from hexbytes import HexBytes
from starknet_py.net.client_models import ContractClass
from starkware.starknet.core.os.class_hash import compute_class_hash
from starkware.starknet.testing.contract_utils import get_contract_class

if TYPE_CHECKING:
    from ape_starknet.accounts import StarknetAccountContainer
    from ape_starknet.config import StarknetConfig
    from ape_starknet.ecosystems import Starknet
    from ape_starknet.explorer import StarknetExplorer
    from ape_starknet.provider import StarknetProvider
    from ape_starknet.tokens import TokenManager
    from ape_starknet.udc import UniversalDeployer


class StarknetBase(ManagerAccessMixin):
    @property
    def starknet_config(self) -> "StarknetConfig":
        return cast("StarknetConfig", self.config_manager.get_config("starknet"))

    @property
    def starknet(self) -> "Starknet":
        return cast("Starknet", self.network_manager.starknet)

    @property
    def starknet_explorer(self) -> "StarknetExplorer":
        explorer = self.provider.network.explorer
        assert explorer  # For mypy
        return cast("StarknetExplorer", explorer)

    @property
    def provider(self) -> "StarknetProvider":
        return cast("StarknetProvider", super().provider)

    @property
    def account_container(self) -> "StarknetAccountContainer":
        return cast("StarknetAccountContainer", self.account_manager.containers["starknet"])

    @property
    def tokens(self) -> "TokenManager":
        from ape_starknet.tokens import tokens

        return tokens

    @cached_property
    def universal_deployer(self) -> "UniversalDeployer":
        from ape_starknet.udc import UniversalDeployer

        return UniversalDeployer()

    def get_contract_type(self, address: AddressType) -> Optional[ContractType]:
        # Force explorer in Local networks since Ape turns it off by default
        return self.chain_manager.contracts.get(
            address
        ) or self.starknet_explorer.get_contract_type(address)

    def get_local_contract_type(self, class_hash: int) -> Optional[ContractType]:
        for contract_name, contract_type in self.project_manager.contracts.items():
            if not contract_type.source_id or not contract_type.source_id.endswith(".cairo"):
                continue

            program = contract_type.deployment_bytecode
            if not program:
                continue

            code = program.bytecode
            if not code:
                continue

            try:
                contract_class = create_contract_class(code)
            except UnicodeDecodeError:
                continue

            contract_cls = get_contract_class(contract_class=contract_class)
            computed_class_hash = compute_class_hash(contract_cls)
            if computed_class_hash == class_hash:
                return contract_type

        return None


def create_contract_class(code: Union[str, bytes]) -> ContractClass:
    if isinstance(code, str) and is_0x_prefixed(code):
        return ContractClass.deserialize(HexBytes(code))

    elif isinstance(code, str):
        return ContractClass.loads(code)

    elif isinstance(code, bytes):
        return ContractClass.deserialize(code)

    else:
        raise TypeError(f"Unhandled bytecode type '{code}'.")



from typing import TYPE_CHECKING, Dict, Union, cast

from ape.api import AccountAPI, Address
from ape.api.networks import LOCAL_NETWORK_NAME
from ape.contracts import ContractInstance
from ape.exceptions import ContractError
from ape.logging import logger
from ape.types import AddressType
from ape.utils import cached_property
from ethpm_types import ContractType
from starknet_devnet.fee_token import FeeToken
from starknet_py.constants import FEE_CONTRACT_ADDRESS

from ape_starknet.ecosystems import NETWORKS
from ape_starknet.exceptions import StarknetTokensError
from ape_starknet.utils import STARKNET_FEE_TOKEN_SYMBOL, to_int
from ape_starknet.utils.basemodel import StarknetBase

if TYPE_CHECKING:
    from ape_starknet.accounts import BaseStarknetAccount


def missing_contract_error(token: str, contract_address: AddressType) -> ContractError:
    return ContractError(f"Incorrect '{token}' contract address '{contract_address}'.")


ERC20 = ContractType(
    **{
        "contractName": "ERC20",
        "abi": [
            {
                "type": "struct",
                "name": "Uint256",
                "members": [
                    {"name": "low", "type": "felt", "offset": 0},
                    {"name": "high", "type": "felt", "offset": 1},
                ],
                "size": 2,
            },
            {
                "type": "event",
                "name": "Transfer",
                "inputs": [
                    {"name": "from_", "type": "felt", "indexed": False},
                    {"name": "to", "type": "felt", "indexed": False},
                    {"name": "value", "type": "Uint256", "indexed": False},
                ],
                "anonymous": False,
            },
            {
                "type": "event",
                "name": "Approval",
                "inputs": [
                    {"name": "owner", "type": "felt", "indexed": False},
                    {"name": "spender", "type": "felt", "indexed": False},
                    {"name": "value", "type": "Uint256", "indexed": False},
                ],
                "anonymous": False,
            },
            {
                "type": "constructor",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "name", "type": "felt"},
                    {"name": "symbol", "type": "felt"},
                    {"name": "decimals", "type": "felt"},
                    {"name": "initial_supply", "type": "Uint256"},
                    {"name": "recipient", "type": "felt"},
                ],
            },
            {
                "type": "function",
                "name": "name",
                "stateMutability": "view",
                "inputs": [],
                "outputs": [{"name": "name", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "symbol",
                "stateMutability": "view",
                "inputs": [],
                "outputs": [{"name": "symbol", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "totalSupply",
                "stateMutability": "view",
                "inputs": [],
                "outputs": [{"name": "totalSupply", "type": "Uint256"}],
            },
            {
                "type": "function",
                "name": "decimals",
                "stateMutability": "view",
                "inputs": [],
                "outputs": [{"name": "decimals", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "balanceOf",
                "stateMutability": "view",
                "inputs": [{"name": "account", "type": "felt"}],
                "outputs": [{"name": "balance", "type": "Uint256"}],
            },
            {
                "type": "function",
                "name": "allowance",
                "stateMutability": "view",
                "inputs": [{"name": "owner", "type": "felt"}, {"name": "spender", "type": "felt"}],
                "outputs": [{"name": "remaining", "type": "Uint256"}],
            },
            {
                "type": "function",
                "name": "transfer",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "recipient", "type": "felt"},
                    {"name": "amount", "type": "Uint256"},
                ],
                "outputs": [{"name": "success", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "transferFrom",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "sender", "type": "felt"},
                    {"name": "recipient", "type": "felt"},
                    {"name": "amount", "type": "Uint256"},
                ],
                "outputs": [{"name": "success", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "approve",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "spender", "type": "felt"},
                    {"name": "amount", "type": "Uint256"},
                ],
                "outputs": [{"name": "success", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "increaseAllowance",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "spender", "type": "felt"},
                    {"name": "added_value", "type": "Uint256"},
                ],
                "outputs": [{"name": "success", "type": "felt"}],
            },
            {
                "type": "function",
                "name": "decreaseAllowance",
                "stateMutability": "nonpayable",
                "inputs": [
                    {"name": "spender", "type": "felt"},
                    {"name": "subtracted_value", "type": "Uint256"},
                ],
                "outputs": [{"name": "success", "type": "felt"}],
            },
        ],
    }
)
TEST_TOKEN_ADDRESS = "0x07394cbe418daa16e42b87ba67372d4ab4a5df0b05c6e554d158458ce245bc10"


class TokenManager(StarknetBase):
    # The 'test_token' refers to the token that comes with Argent-X
    additional_tokens: Dict[str, Dict[str, int]] = {}
    contract_type = ERC20

    # Map of address int to symbols to balance.
    balance_cache: Dict[int, Dict[str, int]] = {}

    cache_enabled: Dict[str, bool] = {LOCAL_NETWORK_NAME: True, **{n: False for n in NETWORKS}}

    @property
    def token_address_map(self) -> Dict:
        return {
            **self._base_token_address_map,
            **self.additional_tokens,
        }

    @cached_property
    def _base_token_address_map(self) -> Dict[str, Dict[str, int]]:
        local_eth = FeeToken.ADDRESS
        live_eth = to_int(FEE_CONTRACT_ADDRESS)
        live_token = to_int(TEST_TOKEN_ADDRESS)

        if self.provider.network.name == LOCAL_NETWORK_NAME:
            self.chain_manager.contracts[
                self.starknet.decode_address(local_eth)
            ] = self.contract_type
        else:
            self.chain_manager.contracts[
                self.starknet.decode_address(live_eth)
            ] = self.contract_type

        return {
            "eth": {
                "local": local_eth,
                "testnet": live_eth,
                "testnet2": live_eth,
                "mainnet": live_eth,
            },
            "test_token": {"testnet": live_token, "testnet2": live_token, "mainnet": live_token},
        }

    def __getitem__(self, token: str) -> ContractInstance:
        network = self.provider.network.name
        token = token.lower()
        if token not in self.token_address_map or not self.token_address_map[token]:
            raise StarknetTokensError(f"Token '{token}' not found.")

        address = self.token_address_map[token].get(network)
        if not address:
            available_networks = ",".join(self.token_address_map[token])
            raise StarknetTokensError(
                f"Token '{token}' not deployed on network "
                f"'{network}' (available networks={available_networks})."
            )

        address_str = self.starknet.decode_address(address)
        return ContractInstance(address_str, ERC20)

    def is_token(self, address: Union[AddressType, int, Address]) -> bool:
        network = self.provider.network.name
        address_int = to_int(address)
        return any(
            address_int == networks.get(network) for networks in self.token_address_map.values()
        )

    def add_token(self, name: str, network: str, address: Union[AddressType, int]):
        if name in self.additional_tokens:
            self.additional_tokens[name][network] = to_int(address)
        else:
            self.additional_tokens[name] = {network: to_int(address)}

    def get_balance(
        self, account: Union[Address, AddressType], token: str = STARKNET_FEE_TOKEN_SYMBOL.lower()
    ) -> int:
        if hasattr(account, "address"):
            address = cast(Address, account).address
        else:
            address = cast(AddressType, account)

        address_int = to_int(address)

        # Al

        network = self.provider.network.name
        if not self.cache_enabled.get(network, False):
            # Strictly use provider.
            balance = self.request_balance(address, token=token)
            self.balance_cache[address_int] = {token: balance}

        elif address_int not in self.balance_cache:
            balance = self.request_balance(address, token=token)
            self.balance_cache[address_int] = {token: balance}

        elif token not in self.balance_cache[address_int]:
            self.balance_cache[address_int][token] = self.request_balance(address, token=token)

        return self.balance_cache[address_int][token]

    def request_balance(
        self, account: Union[AddressType, int], token: str = STARKNET_FEE_TOKEN_SYMBOL.lower()
    ) -> int:
        """
        Get the balance from the provider and update the cache.
        """

        account_int = to_int(account)
        amount = self[token].balanceOf(account_int)
        amount_int = self._convert_amount_to_int(amount)

        # Update cache to save requests (only if caching enabled).
        if account_int in self.balance_cache:
            self.balance_cache[account_int][token] = amount_int
        else:
            self.balance_cache[account_int] = {token: amount_int}

        return amount_int

    def transfer(
        self,
        sender: Union[int, AddressType, "BaseStarknetAccount"],
        receiver: Union[int, AddressType, "BaseStarknetAccount"],
        amount: int,
        token: str = STARKNET_FEE_TOKEN_SYMBOL.lower(),
    ):
        receiver_int = to_int(receiver)
        sender_account = cast(
            "BaseStarknetAccount",
            (self.account_container[sender] if isinstance(sender, (int, str)) else sender),
        )
        result = self[token].transfer(receiver_int, amount, sender=sender_account)

        # NOTE: the fees paid by the sender get updated in `provider.send_transaction()`.
        amount_int = self._convert_amount_to_int(amount)
        self.update_cache(sender_account.address_int, -amount_int, token=token)
        self.update_cache(receiver_int, amount_int, token=token)
        return result

    def update_cache(
        self,
        address: Union[AccountAPI, AddressType, int],
        amount: Union[int, Dict],
        token: str = STARKNET_FEE_TOKEN_SYMBOL.lower(),
    ):
        amount_int = self._convert_amount_to_int(amount)
        address_int = to_int(address)
        if address_int not in self.balance_cache or token not in self.balance_cache[address_int]:
            # Set the balance from a request to the provider.
            self.request_balance(address_int, token=token)
            return

        current_balance: int = self.balance_cache[address_int][token]
        if current_balance + amount_int < 0:
            actual_balance = self.request_balance(address_int, token=token)
            logger.error(
                f"Balance cache corrupted - "
                f"attempted to set as {amount_int} when actual balance is {actual_balance}"
            )

        else:
            self.balance_cache[address_int][token] += amount_int

    def _convert_amount_to_int(self, amount: Union[int, Dict]) -> int:
        if isinstance(amount, int):
            return amount

        elif isinstance(amount, dict) and "low" in amount and "high" in amount:
            return (amount["high"] << 128) + amount["low"]

        elif isinstance(amount, dict) and "low" in amount:
            return amount["low"]

        elif isinstance(amount, (list, tuple)) and len(amount) == 2:
            return (amount[1] << 128) + amount[0]

        elif isinstance(amount, (list, tuple)) and len(amount) == 1:
            return amount[0]

        raise StarknetTokensError(f"Unable to handle transfer value '{amount}'.")


tokens = TokenManager()



from typing import Any, Dict, List, Optional, cast

from ape.types import AddressType
from ape.utils import cached_property
from ethpm_types import ContractType
from ethpm_types.abi import MethodABI
from starkware.starknet.definitions.fields import ContractAddressSalt

from ape_starknet.transactions import InvokeFunctionTransaction
from ape_starknet.utils.basemodel import StarknetBase

DEFAULT_UDC_ADDRESS: AddressType = cast(
    AddressType, "0x041A78e741e5aF2fec34B695679bc6891742439F7Afb8484ecd7766661aD02bF"
)
DEFAULT_UDC_ABI: List[Dict] = [
    {
        "type": "event",
        "name": "ContractDeployed",
        "inputs": [
            {"name": "address", "type": "felt", "indexed": False},
            {"name": "deployer", "type": "felt", "indexed": False},
            {"name": "unique", "type": "felt", "indexed": False},
            {"name": "classHash", "type": "felt", "indexed": False},
            {"name": "calldata_len", "type": "felt", "indexed": False},
            {"name": "calldata", "type": "felt*", "indexed": False},
            {"name": "salt", "type": "felt", "indexed": False},
        ],
        "anonymous": False,
    },
    {
        "type": "function",
        "name": "deployContract",
        "stateMutability": "nonpayable",
        "inputs": [
            {"name": "classHash", "type": "felt"},
            {"name": "salt", "type": "felt"},
            {"name": "unique", "type": "felt"},
            {"name": "calldata_len", "type": "felt"},
            {"name": "calldata", "type": "felt*"},
        ],
        "outputs": [{"name": "address", "type": "felt"}],
    },
]


class UniversalDeployer(StarknetBase):
    @cached_property
    def contract_type(self) -> ContractType:
        return ContractType(
            contractName="openzeppelin.utils.presets.UniversalDeployer",
            sourceId="openzeppelin/utils/presets/UniversalDeployer.cairo",
            abi=self.abi,
            # NOTE: code is not necessary for its use-cases (use like an interface)
            deploymentBytecode={},
            runtimeBytecode={},
        )

    def __init__(
        self, address: AddressType = DEFAULT_UDC_ADDRESS, abi: Optional[List[Dict]] = None
    ):
        self.address = address
        self.abi = abi or DEFAULT_UDC_ABI
        self._cache_self()
        super().__init__()

    @cached_property
    def deploy_function(self) -> MethodABI:
        method = self.contract_type.mutable_methods["deployContract"]
        method.contract_type = self.contract_type
        return method

    def create_deploy(
        self,
        class_hash: int,
        constructor_arguments: List[Any],
        salt: Optional[int] = None,
        unique: bool = True,
        **kwargs,
    ) -> InvokeFunctionTransaction:
        salt = salt or ContractAddressSalt.get_random_value()
        txn = self.starknet.encode_transaction(
            self.address,
            self.deploy_function,
            class_hash,
            salt,
            unique,
            len(constructor_arguments),
            constructor_arguments,
            **kwargs,
        )
        return cast(InvokeFunctionTransaction, txn)

    def _cache_self(self):
        self.chain_manager.contracts[self.address] = self.contract_type



import pytest
from ape.contracts import ContractContainer, ContractInstance


@pytest.fixture
def in_ethereum(use_local_ethereum):
    with use_local_ethereum:
        yield


@pytest.fixture
def in_starknet(use_local_starknet):
    with use_local_starknet:
        yield


@pytest.fixture(scope="module")
def eth_contract_container(eth_contract_type):
    return ContractContainer(eth_contract_type)


@pytest.fixture(scope="module", autouse=True)
def deploy_eth_contract(use_local_ethereum, eth_account, eth_contract_container):
    with use_local_ethereum:
        return eth_account.deploy(eth_contract_container, sender=eth_account)


@pytest.fixture
def eth_contract(in_ethereum, eth_contract_container):
    return eth_contract_container.deployments[-1]


@pytest.fixture(scope="module")
def stark_contract(use_local_starknet, config, project_path):
    with use_local_starknet:
        with config.using_project(project_path) as project:
            yield project.MyContract.deployments[-1]


def test_use_eth_network_from_fixture(eth_contract, eth_account, in_ethereum):
    # Shows that we can write Ethereum-only tests within a multi-chain test module
    # (NOTE: 'starknet' is the default ecosystem for this project)
    eth_contract.setNumber(123, sender=eth_account)
    assert eth_contract.myNumber() == 123


def test_use_starknet_network_from_fixture(account, stark_contract, in_starknet):
    # Shows that we can write Starknet-only tests within a multi-chain test module
    receipt = stark_contract.increase_balance(account.address, 123, sender=account)
    assert not receipt.failed


def test_switch_to_ethereum_mid_test(
    provider,
    eth_account,
    eth_contract_container,
    stark_contract,
    account,
    in_starknet,
    use_local_ethereum,
):
    receipt = stark_contract.increase_balance(account.address, 123, sender=account)
    assert not receipt.failed

    # Shows that we can change to Ethereum within an individual test
    starknet_chain_id = provider.chain_id

    with use_local_ethereum as eth_provider:
        # Verify the chain changed
        assert eth_provider.chain_id != starknet_chain_id

        # Deploy and interact with a contract on Ethereum
        eth_contract = eth_contract_container.deploy(sender=eth_account)
        receipt = eth_contract.setNumber(123, sender=eth_account)
        assert not receipt.failed
        assert isinstance(eth_contract, ContractInstance)
        assert eth_contract.myNumber() == 123

    # Switch back to Starknet
    receipt = stark_contract.increase_balance(account.address, 123, sender=account)
    assert not receipt.failed
